from __future__ import absolute_import
from __future__ import print_function

import os
import veriloggen
import thread_axi_stream

expected_verilog = """
module blinkled
(
  input CLK,
  input RST,
  input [32-1:0] axi_a_tdata,
  input axi_a_tvalid,
  output axi_a_tready,
  input axi_a_tlast,
  output reg [32-1:0] axi_b_tdata,
  output reg axi_b_tvalid,
  input axi_b_tready,
  output reg axi_b_tlast,
  input [32-1:0] saxi_awaddr,
  input [4-1:0] saxi_awcache,
  input [3-1:0] saxi_awprot,
  input saxi_awvalid,
  output saxi_awready,
  input [32-1:0] saxi_wdata,
  input [4-1:0] saxi_wstrb,
  input saxi_wvalid,
  output saxi_wready,
  output [2-1:0] saxi_bresp,
  output reg saxi_bvalid,
  input saxi_bready,
  input [32-1:0] saxi_araddr,
  input [4-1:0] saxi_arcache,
  input [3-1:0] saxi_arprot,
  input saxi_arvalid,
  output saxi_arready,
  output reg [32-1:0] saxi_rdata,
  output [2-1:0] saxi_rresp,
  output reg saxi_rvalid,
  input saxi_rready
);

  reg _axi_a_read_start;
  reg [8-1:0] _axi_a_read_op_sel;
  reg [32-1:0] _axi_a_read_local_addr;
  reg [33-1:0] _axi_a_read_size;
  reg [32-1:0] _axi_a_read_local_stride;
  reg _axi_a_read_idle;
  reg _axi_b_write_start;
  reg [8-1:0] _axi_b_write_op_sel;
  reg [32-1:0] _axi_b_write_local_addr;
  reg [33-1:0] _axi_b_write_size;
  reg [32-1:0] _axi_b_write_local_stride;
  reg _axi_b_write_idle;
  assign saxi_bresp = 0;
  assign saxi_rresp = 0;
  reg signed [32-1:0] _saxi_register_0;
  reg signed [32-1:0] _saxi_register_1;
  reg signed [32-1:0] _saxi_register_2;
  reg signed [32-1:0] _saxi_register_3;
  reg _saxi_flag_0;
  reg _saxi_flag_1;
  reg _saxi_flag_2;
  reg _saxi_flag_3;
  reg signed [32-1:0] _saxi_resetval_0;
  reg signed [32-1:0] _saxi_resetval_1;
  reg signed [32-1:0] _saxi_resetval_2;
  reg signed [32-1:0] _saxi_resetval_3;
  localparam _saxi_maskwidth = 2;
  localparam _saxi_mask = { _saxi_maskwidth{ 1'd1 } };
  localparam _saxi_shift = 2;
  reg [32-1:0] _saxi_register_fsm;
  localparam _saxi_register_fsm_init = 0;
  reg [32-1:0] _tmp_0;
  reg _tmp_1;
  reg _tmp_2;
  reg _tmp_3;
  reg _tmp_4;
  assign saxi_awready = (_saxi_register_fsm == 0) && !_tmp_1 && !_tmp_2 && !saxi_bvalid && _tmp_3;
  assign saxi_arready = (_saxi_register_fsm == 0) && !_tmp_2 && !_tmp_1 && _tmp_4;
  reg [_saxi_maskwidth-1:0] _tmp_5;
  wire signed [32-1:0] _tmp_6;
  assign _tmp_6 = (_tmp_5 == 0)? _saxi_register_0 : 
                  (_tmp_5 == 1)? _saxi_register_1 : 
                  (_tmp_5 == 2)? _saxi_register_2 : 
                  (_tmp_5 == 3)? _saxi_register_3 : 'hx;
  wire _tmp_7;
  assign _tmp_7 = (_tmp_5 == 0)? _saxi_flag_0 : 
                  (_tmp_5 == 1)? _saxi_flag_1 : 
                  (_tmp_5 == 2)? _saxi_flag_2 : 
                  (_tmp_5 == 3)? _saxi_flag_3 : 'hx;
  wire signed [32-1:0] _tmp_8;
  assign _tmp_8 = (_tmp_5 == 0)? _saxi_resetval_0 : 
                  (_tmp_5 == 1)? _saxi_resetval_1 : 
                  (_tmp_5 == 2)? _saxi_resetval_2 : 
                  (_tmp_5 == 3)? _saxi_resetval_3 : 'hx;
  reg _saxi_cond_0_1;
  assign saxi_wready = _saxi_register_fsm == 3;
  reg [32-1:0] th_comp;
  localparam th_comp_init = 0;
  reg signed [32-1:0] _th_comp_size_0;
  reg signed [32-1:0] _th_comp_i_1;
  assign axi_a_tready = th_comp == 7;
  reg signed [32-1:0] axistreamin_rdata_9;
  reg [1-1:0] axistreamin_rlast_10;
  reg signed [32-1:0] _th_comp_a_2;
  reg signed [32-1:0] _th_comp_a_last_3;
  reg signed [32-1:0] _th_comp_b_4;
  reg signed [32-1:0] _th_comp_b_last_5;
  reg _axi_b_cond_0_1;

  always @(posedge CLK) begin
    if(RST) begin
      _axi_a_read_start <= 0;
    end else begin
      _axi_a_read_start <= 0;
    end
  end


  always @(posedge CLK) begin
    if(RST) begin
      _axi_b_write_start <= 0;
      axi_b_tdata <= 0;
      axi_b_tvalid <= 0;
      axi_b_tlast <= 0;
      _axi_b_cond_0_1 <= 0;
    end else begin
      if(_axi_b_cond_0_1) begin
        axi_b_tvalid <= 0;
        axi_b_tlast <= 0;
      end 
      _axi_b_write_start <= 0;
      if((th_comp == 11) && (axi_b_tready || !axi_b_tvalid)) begin
        axi_b_tdata <= _th_comp_b_4;
        axi_b_tvalid <= 1;
        axi_b_tlast <= _th_comp_b_last_5;
      end 
      _axi_b_cond_0_1 <= 1;
      if(axi_b_tvalid && !axi_b_tready) begin
        axi_b_tvalid <= axi_b_tvalid;
        axi_b_tlast <= axi_b_tlast;
      end 
    end
  end


  always @(posedge CLK) begin
    if(RST) begin
      saxi_bvalid <= 0;
      _tmp_3 <= 0;
      _tmp_4 <= 0;
      _tmp_1 <= 0;
      _tmp_2 <= 0;
      _tmp_0 <= 0;
      saxi_rdata <= 0;
      saxi_rvalid <= 0;
      _saxi_cond_0_1 <= 0;
      _saxi_register_0 <= 0;
      _saxi_flag_0 <= 0;
      _saxi_register_1 <= 0;
      _saxi_flag_1 <= 0;
      _saxi_register_2 <= 0;
      _saxi_flag_2 <= 0;
      _saxi_register_3 <= 0;
      _saxi_flag_3 <= 0;
    end else begin
      if(_saxi_cond_0_1) begin
        saxi_rvalid <= 0;
      end 
      if(saxi_bvalid && saxi_bready) begin
        saxi_bvalid <= 0;
      end 
      if(saxi_wvalid && saxi_wready) begin
        saxi_bvalid <= 1;
      end 
      _tmp_3 <= saxi_awvalid;
      _tmp_4 <= saxi_arvalid;
      _tmp_1 <= 0;
      _tmp_2 <= 0;
      if(saxi_awready && saxi_awvalid && !saxi_bvalid) begin
        _tmp_0 <= saxi_awaddr;
        _tmp_1 <= 1;
      end else if(saxi_arready && saxi_arvalid) begin
        _tmp_0 <= saxi_araddr;
        _tmp_2 <= 1;
      end 
      if((_saxi_register_fsm == 1) && (saxi_rready || !saxi_rvalid)) begin
        saxi_rdata <= _tmp_6;
        saxi_rvalid <= 1;
      end 
      _saxi_cond_0_1 <= 1;
      if(saxi_rvalid && !saxi_rready) begin
        saxi_rvalid <= saxi_rvalid;
      end 
      if((_saxi_register_fsm == 1) && (saxi_rready || !saxi_rvalid) && _tmp_7 && (_tmp_5 == 0)) begin
        _saxi_register_0 <= _tmp_8;
        _saxi_flag_0 <= 0;
      end 
      if((_saxi_register_fsm == 1) && (saxi_rready || !saxi_rvalid) && _tmp_7 && (_tmp_5 == 1)) begin
        _saxi_register_1 <= _tmp_8;
        _saxi_flag_1 <= 0;
      end 
      if((_saxi_register_fsm == 1) && (saxi_rready || !saxi_rvalid) && _tmp_7 && (_tmp_5 == 2)) begin
        _saxi_register_2 <= _tmp_8;
        _saxi_flag_2 <= 0;
      end 
      if((_saxi_register_fsm == 1) && (saxi_rready || !saxi_rvalid) && _tmp_7 && (_tmp_5 == 3)) begin
        _saxi_register_3 <= _tmp_8;
        _saxi_flag_3 <= 0;
      end 
      if((_saxi_register_fsm == 3) && (saxi_wready && saxi_wvalid) && (_tmp_5 == 0)) begin
        _saxi_register_0 <= saxi_wdata;
      end 
      if((_saxi_register_fsm == 3) && (saxi_wready && saxi_wvalid) && (_tmp_5 == 1)) begin
        _saxi_register_1 <= saxi_wdata;
      end 
      if((_saxi_register_fsm == 3) && (saxi_wready && saxi_wvalid) && (_tmp_5 == 2)) begin
        _saxi_register_2 <= saxi_wdata;
      end 
      if((_saxi_register_fsm == 3) && (saxi_wready && saxi_wvalid) && (_tmp_5 == 3)) begin
        _saxi_register_3 <= saxi_wdata;
      end 
      if((_saxi_register_0 == 1) && (th_comp == 2) && 1) begin
        _saxi_register_0 <= 0;
      end 
      if((_saxi_register_0 == 1) && (th_comp == 2) && 0) begin
        _saxi_register_1 <= 0;
      end 
      if((_saxi_register_0 == 1) && (th_comp == 2) && 0) begin
        _saxi_register_2 <= 0;
      end 
      if((_saxi_register_0 == 1) && (th_comp == 2) && 0) begin
        _saxi_register_3 <= 0;
      end 
      if((th_comp == 3) && 0) begin
        _saxi_register_0 <= 1;
        _saxi_flag_0 <= 0;
      end 
      if((th_comp == 3) && 1) begin
        _saxi_register_1 <= 1;
        _saxi_flag_1 <= 0;
      end 
      if((th_comp == 3) && 0) begin
        _saxi_register_2 <= 1;
        _saxi_flag_2 <= 0;
      end 
      if((th_comp == 3) && 0) begin
        _saxi_register_3 <= 1;
        _saxi_flag_3 <= 0;
      end 
      if((th_comp == 13) && 0) begin
        _saxi_register_0 <= 0;
        _saxi_flag_0 <= 0;
      end 
      if((th_comp == 13) && 1) begin
        _saxi_register_1 <= 0;
        _saxi_flag_1 <= 0;
      end 
      if((th_comp == 13) && 0) begin
        _saxi_register_2 <= 0;
        _saxi_flag_2 <= 0;
      end 
      if((th_comp == 13) && 0) begin
        _saxi_register_3 <= 0;
        _saxi_flag_3 <= 0;
      end 
    end
  end

  localparam _saxi_register_fsm_1 = 1;
  localparam _saxi_register_fsm_2 = 2;
  localparam _saxi_register_fsm_3 = 3;

  always @(posedge CLK) begin
    if(RST) begin
      _saxi_register_fsm <= _saxi_register_fsm_init;
    end else begin
      case(_saxi_register_fsm)
        _saxi_register_fsm_init: begin
          if(_tmp_2 || _tmp_1) begin
            _tmp_5 <= (_tmp_0 >> _saxi_shift) & _saxi_mask;
          end 
          if(_tmp_2) begin
            _saxi_register_fsm <= _saxi_register_fsm_1;
          end 
          if(_tmp_1) begin
            _saxi_register_fsm <= _saxi_register_fsm_3;
          end 
        end
        _saxi_register_fsm_1: begin
          if(saxi_rready && saxi_rvalid) begin
            _saxi_register_fsm <= _saxi_register_fsm_init;
          end 
          if((saxi_rready || !saxi_rvalid) && !(saxi_rready && saxi_rvalid)) begin
            _saxi_register_fsm <= _saxi_register_fsm_2;
          end 
        end
        _saxi_register_fsm_2: begin
          if(saxi_rready && saxi_rvalid) begin
            _saxi_register_fsm <= _saxi_register_fsm_init;
          end 
        end
        _saxi_register_fsm_3: begin
          if(saxi_wready && saxi_wvalid) begin
            _saxi_register_fsm <= _saxi_register_fsm_init;
          end 
        end
      endcase
    end
  end

  localparam th_comp_1 = 1;
  localparam th_comp_2 = 2;
  localparam th_comp_3 = 3;
  localparam th_comp_4 = 4;
  localparam th_comp_5 = 5;
  localparam th_comp_6 = 6;
  localparam th_comp_7 = 7;
  localparam th_comp_8 = 8;
  localparam th_comp_9 = 9;
  localparam th_comp_10 = 10;
  localparam th_comp_11 = 11;
  localparam th_comp_12 = 12;
  localparam th_comp_13 = 13;
  localparam th_comp_14 = 14;
  localparam th_comp_15 = 15;
  localparam th_comp_16 = 16;

  always @(posedge CLK) begin
    if(RST) begin
      th_comp <= th_comp_init;
      _th_comp_size_0 <= 0;
      _th_comp_i_1 <= 0;
      axistreamin_rdata_9 <= 0;
      axistreamin_rlast_10 <= 0;
      _th_comp_a_2 <= 0;
      _th_comp_a_last_3 <= 0;
      _th_comp_b_4 <= 0;
      _th_comp_b_last_5 <= 0;
    end else begin
      case(th_comp)
        th_comp_init: begin
          th_comp <= th_comp_1;
        end
        th_comp_1: begin
          if(1) begin
            th_comp <= th_comp_2;
          end else begin
            th_comp <= th_comp_15;
          end
        end
        th_comp_2: begin
          if(_saxi_register_0 == 1) begin
            th_comp <= th_comp_3;
          end 
        end
        th_comp_3: begin
          th_comp <= th_comp_4;
        end
        th_comp_4: begin
          _th_comp_size_0 <= _saxi_register_2;
          th_comp <= th_comp_5;
        end
        th_comp_5: begin
          _th_comp_i_1 <= 0;
          th_comp <= th_comp_6;
        end
        th_comp_6: begin
          if(_th_comp_i_1 < _th_comp_size_0) begin
            th_comp <= th_comp_7;
          end else begin
            th_comp <= th_comp_13;
          end
        end
        th_comp_7: begin
          if(axi_a_tready && axi_a_tvalid) begin
            axistreamin_rdata_9 <= axi_a_tdata;
            axistreamin_rlast_10 <= axi_a_tlast;
          end 
          if(axi_a_tready && axi_a_tvalid) begin
            th_comp <= th_comp_8;
          end 
        end
        th_comp_8: begin
          _th_comp_a_2 <= axistreamin_rdata_9;
          _th_comp_a_last_3 <= axistreamin_rlast_10;
          th_comp <= th_comp_9;
        end
        th_comp_9: begin
          _th_comp_b_4 <= _th_comp_a_2 + 1;
          th_comp <= th_comp_10;
        end
        th_comp_10: begin
          _th_comp_b_last_5 <= _th_comp_a_last_3;
          th_comp <= th_comp_11;
        end
        th_comp_11: begin
          if(axi_b_tready || !axi_b_tvalid) begin
            th_comp <= th_comp_12;
          end 
        end
        th_comp_12: begin
          _th_comp_i_1 <= _th_comp_i_1 + 1;
          th_comp <= th_comp_6;
        end
        th_comp_13: begin
          th_comp <= th_comp_14;
        end
        th_comp_14: begin
          th_comp <= th_comp_1;
        end
        th_comp_15: begin
          $finish;
          th_comp <= th_comp_16;
        end
      endcase
    end
  end


endmodule
"""


def test(request):
    veriloggen.reset()

    simtype = request.config.getoption('--sim')

    code = thread_axi_stream.run(filename=None, simtype=simtype,
                                 outputfile=os.path.splitext(os.path.basename(__file__))[0] + '.out')

    from pyverilog.vparser.parser import VerilogParser
    from pyverilog.ast_code_generator.codegen import ASTCodeGenerator

    parser = VerilogParser()
    expected_ast = parser.parse(expected_verilog)
    codegen = ASTCodeGenerator()
    expected_code = codegen.visit(expected_ast)

    assert(expected_code == code)



from __future__ import absolute_import
from __future__ import print_function
import veriloggen
import stream_substream

expected_verilog = """
module test;

  reg CLK;
  reg RST;
  reg signed [32-1:0] xdata;
  reg signed [32-1:0] ydata;
  reg signed [32-1:0] edata;
  wire signed [32-1:0] zdata;
  wire [1-1:0] vdata;

  main
  uut
  (
    .CLK(CLK),
    .RST(RST),
    .xdata(xdata),
    .ydata(ydata),
    .edata(edata),
    .zdata(zdata),
    .vdata(vdata)
  );

  reg reset_done;

  initial begin
    $dumpfile("uut.vcd");
    $dumpvars(0, uut);
  end


  initial begin
    CLK = 0;
    forever begin
      #5 CLK = !CLK;
    end
  end


  initial begin
    RST = 0;
    reset_done = 0;
    xdata = 0;
    ydata = 0;
    edata = 0;
    #100;
    RST = 1;
    #100;
    RST = 0;
    #1000;
    reset_done = 1;
    @(posedge CLK);
    #1;
    #10000;
    $finish;
  end

  reg [32-1:0] send_fsm;
  localparam send_fsm_init = 0;
  reg [32-1:0] send_count;
  reg [32-1:0] _d1_send_fsm;
  reg _send_fsm_cond_1_0_1;
  reg _send_fsm_cond_2_1_1;
  reg [32-1:0] recv_fsm;
  localparam recv_fsm_init = 0;
  reg [32-1:0] recv_count;
  localparam send_fsm_1 = 1;
  localparam send_fsm_2 = 2;
  localparam send_fsm_3 = 3;

  always @(posedge CLK) begin
    if(RST) begin
      send_fsm <= send_fsm_init;
      _d1_send_fsm <= send_fsm_init;
      send_count <= 0;
      _send_fsm_cond_1_0_1 <= 0;
      _send_fsm_cond_2_1_1 <= 0;
    end else begin
      _d1_send_fsm <= send_fsm;
      case(_d1_send_fsm)
        send_fsm_1: begin
          if(_send_fsm_cond_1_0_1) begin
            $display("xdata=%d", xdata);
            $display("ydata=%d", ydata);
          end 
        end
        send_fsm_2: begin
          if(_send_fsm_cond_2_1_1) begin
            $display("xdata=%d", xdata);
            $display("ydata=%d", ydata);
          end 
        end
      endcase
      case(send_fsm)
        send_fsm_init: begin
          if(reset_done) begin
            send_fsm <= send_fsm_1;
          end 
        end
        send_fsm_1: begin
          xdata <= 0;
          ydata <= 0;
          edata <= 1;
          send_count <= send_count + 1;
          _send_fsm_cond_1_0_1 <= 1;
          send_fsm <= send_fsm_2;
        end
        send_fsm_2: begin
          xdata <= xdata + 1;
          ydata <= ydata + 2;
          edata <= 1;
          send_count <= send_count + 1;
          _send_fsm_cond_2_1_1 <= 1;
          if(send_count == 64) begin
            send_fsm <= send_fsm_3;
          end 
        end
        send_fsm_3: begin
          edata <= 0;
        end
      endcase
    end
  end

  localparam recv_fsm_1 = 1;
  localparam recv_fsm_2 = 2;

  always @(posedge CLK) begin
    if(RST) begin
      recv_fsm <= recv_fsm_init;
      recv_count <= 0;
    end else begin
      case(recv_fsm)
        recv_fsm_init: begin
          if(reset_done) begin
            recv_fsm <= recv_fsm_1;
          end 
        end
        recv_fsm_1: begin
          if(vdata) begin
            $display("zdata=%d", zdata);
            recv_count <= recv_count + 1;
          end 
          if(recv_count == 8) begin
            recv_fsm <= recv_fsm_2;
          end 
        end
      endcase
    end
  end


endmodule



module main
(
  input CLK,
  input RST,
  input signed [32-1:0] xdata,
  input signed [32-1:0] ydata,
  input signed [32-1:0] edata,
  output signed [32-1:0] zdata,
  output [1-1:0] vdata
);

  wire signed [32-1:0] adata;
  wire signed [32-1:0] bdata;
  wire signed [64-1:0] _times_mul_odata_2;
  reg signed [64-1:0] _times_mul_odata_reg_2;
  wire signed [32-1:0] _times_data_2;
  assign _times_data_2 = _times_mul_odata_reg_2;
  wire _times_mul_update_2;
  assign _times_mul_update_2 = 1'd1;

  multiplier_0
  _times_mul_2
  (
    .CLK(CLK),
    .update(_times_mul_update_2),
    .a(adata),
    .b(bdata),
    .c(_times_mul_odata_2)
  );

  wire signed [32-1:0] cdata;
  assign cdata = _times_data_2;
  reg signed [32-1:0] _plus_data_6;
  reg signed [32-1:0] _plus_data_8;
  reg signed [32-1:0] __delay_data_20__variable_5;
  reg signed [32-1:0] __variable_wdata_0;
  assign adata = __variable_wdata_0;
  reg signed [32-1:0] __variable_wdata_1;
  assign bdata = __variable_wdata_1;
  reg signed [32-1:0] __delay_data_21__delay_20__variable_5;
  reg signed [32-1:0] __delay_data_22__delay_21__delay_20__variable_5;
  reg signed [32-1:0] __delay_data_23__delay_22__delay_21__delay_20__variable_5;
  reg signed [32-1:0] __delay_data_24__delay_23__delay_22__delay_21____variable_5;
  wire signed [32-1:0] __substreamoutput_data_11;
  assign __substreamoutput_data_11 = cdata;
  reg signed [32-1:0] _reduceadd_data_12;
  reg [6-1:0] _reduceadd_count_12;
  reg _reduceadd_prev_count_max_12;
  wire _reduceadd_reset_cond_12;
  assign _reduceadd_reset_cond_12 = _reduceadd_prev_count_max_12;
  wire [6-1:0] _reduceadd_current_count_12;
  assign _reduceadd_current_count_12 = (_reduceadd_reset_cond_12)? 0 : _reduceadd_count_12;
  wire signed [32-1:0] _reduceadd_current_data_12;
  assign _reduceadd_current_data_12 = (_reduceadd_reset_cond_12)? 1'sd0 : _reduceadd_data_12;
  reg [1-1:0] _pulse_data_15;
  reg [6-1:0] _pulse_count_15;
  reg _pulse_prev_count_max_15;
  wire _pulse_reset_cond_15;
  assign _pulse_reset_cond_15 = _pulse_prev_count_max_15;
  wire [6-1:0] _pulse_current_count_15;
  assign _pulse_current_count_15 = (_pulse_reset_cond_15)? 0 : _pulse_count_15;
  wire [1-1:0] _pulse_current_data_15;
  assign _pulse_current_data_15 = (_pulse_reset_cond_15)? 1'sd0 : _pulse_data_15;
  reg signed [32-1:0] _plus_data_18;
  reg [1-1:0] __delay_data_25_pulse_15;
  assign zdata = _plus_data_18;
  assign vdata = __delay_data_25_pulse_15;

  always @(posedge CLK) begin
    if(RST) begin
      _times_mul_odata_reg_2 <= 0;
      __variable_wdata_0 <= 0;
      __variable_wdata_1 <= 0;
    end else begin
      _times_mul_odata_reg_2 <= _times_mul_odata_2;
      __variable_wdata_0 <= _plus_data_6;
      __variable_wdata_1 <= _plus_data_8;
    end
  end


  always @(posedge CLK) begin
    if(RST) begin
      _plus_data_6 <= 0;
      _plus_data_8 <= 0;
      __delay_data_20__variable_5 <= 0;
      __delay_data_21__delay_20__variable_5 <= 0;
      __delay_data_22__delay_21__delay_20__variable_5 <= 0;
      __delay_data_23__delay_22__delay_21__delay_20__variable_5 <= 0;
      __delay_data_24__delay_23__delay_22__delay_21____variable_5 <= 0;
      _reduceadd_count_12 <= 0;
      _reduceadd_prev_count_max_12 <= 0;
      _reduceadd_data_12 <= 1'sd0;
      _pulse_count_15 <= 0;
      _pulse_prev_count_max_15 <= 0;
      _pulse_data_15 <= 1'sd0;
      _plus_data_18 <= 0;
      __delay_data_25_pulse_15 <= 0;
    end else begin
      _plus_data_6 <= xdata + 2'sd1;
      _plus_data_8 <= ydata + 3'sd2;
      __delay_data_20__variable_5 <= edata;
      __delay_data_21__delay_20__variable_5 <= __delay_data_20__variable_5;
      __delay_data_22__delay_21__delay_20__variable_5 <= __delay_data_21__delay_20__variable_5;
      __delay_data_23__delay_22__delay_21__delay_20__variable_5 <= __delay_data_22__delay_21__delay_20__variable_5;
      __delay_data_24__delay_23__delay_22__delay_21____variable_5 <= __delay_data_23__delay_22__delay_21__delay_20__variable_5;
      if(__delay_data_24__delay_23__delay_22__delay_21____variable_5) begin
        _reduceadd_count_12 <= (_reduceadd_current_count_12 >= 5'sd8 - 1)? 0 : _reduceadd_current_count_12 + 1;
      end 
      if(__delay_data_24__delay_23__delay_22__delay_21____variable_5) begin
        _reduceadd_prev_count_max_12 <= _reduceadd_current_count_12 >= 5'sd8 - 1;
      end 
      if(__delay_data_24__delay_23__delay_22__delay_21____variable_5) begin
        _reduceadd_data_12 <= _reduceadd_current_data_12 + __substreamoutput_data_11;
      end 
      if(__delay_data_24__delay_23__delay_22__delay_21____variable_5) begin
        _pulse_count_15 <= (_pulse_current_count_15 >= 5'sd8 - 1)? 0 : _pulse_current_count_15 + 1;
      end 
      if(__delay_data_24__delay_23__delay_22__delay_21____variable_5) begin
        _pulse_prev_count_max_15 <= _pulse_current_count_15 >= 5'sd8 - 1;
      end 
      if(__delay_data_24__delay_23__delay_22__delay_21____variable_5) begin
        _pulse_data_15 <= _pulse_current_count_15 >= 5'sd8 - 1;
      end 
      _plus_data_18 <= _reduceadd_data_12 + 11'sd1000;
      __delay_data_25_pulse_15 <= _pulse_data_15;
    end
  end


endmodule



module multiplier_0
(
  input CLK,
  input update,
  input [32-1:0] a,
  input [32-1:0] b,
  output [64-1:0] c
);


  multiplier_core_0
  mult
  (
    .CLK(CLK),
    .update(update),
    .a(a),
    .b(b),
    .c(c)
  );


endmodule



module multiplier_core_0
(
  input CLK,
  input update,
  input [32-1:0] a,
  input [32-1:0] b,
  output [64-1:0] c
);

  reg signed [32-1:0] _a;
  reg signed [32-1:0] _b;
  wire signed [64-1:0] _mul;
  reg signed [64-1:0] _pipe_mul0;
  assign _mul = _a * _b;
  assign c = _pipe_mul0;

  always @(posedge CLK) begin
    if(update) begin
      _a <= a;
      _b <= b;
      _pipe_mul0 <= _mul;
    end 
  end


endmodule
"""


def test():
    veriloggen.reset()
    test_module = stream_substream.mkTest()
    code = test_module.to_verilog()

    from pyverilog.vparser.parser import VerilogParser
    from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
    parser = VerilogParser()
    expected_ast = parser.parse(expected_verilog)
    codegen = ASTCodeGenerator()
    expected_code = codegen.visit(expected_ast)

    assert(expected_code == code)



from __future__ import absolute_import
from __future__ import print_function

import veriloggen.core.vtypes as vtypes
import copy


def add(left, right, width=8, max=None, min=None):
    raw_val = left + right
    return _add(raw_val, left, right, width, max, min)


def sub(left, right, width=8, max=None, min=None):
    raw_val = left - right
    return _sub(raw_val, left, right, width, max, min)


def to_saturate(value, width=8, max=None, min=None, signed=False):
    value = _saturate_op(value, width, max, min, signed)
    return _saturate_single(value, width, max, min, signed)


def _saturate_op(value, width=8, max=None, min=None, signed=False):
    if hasattr(value, 'saturated') and value.saturated:
        return value

    if isinstance(value, vtypes.Plus):
        new_val = copy.deepcopy(value)
        new_val.left = _saturate_op(value.left, width=width, max=max, min=min)
        new_val.right = _saturate_op(value.right,
                                     width=width, max=max, min=min)
        ret = _add(new_val, new_val.left, new_val.right,
                   width=width, max=max, min=min)
        ret.saturated = True
        return ret

    if isinstance(value, vtypes.Minus):
        new_val = copy.deepcopy(value)
        new_val.left = _saturate_op(value.left, width, max)
        new_val.right = _saturate_op(value.right, width, max)
        ret = _sub(new_val, new_val.left, new_val.right,
                   width=width, max=max, min=min)
        ret.saturated = True
        return ret

    return value


def _saturate_single(value, width=8, max=None, min=None, signed=False):
    if hasattr(value, 'saturated') and value.saturated:
        return value

    if (signed or
            (isinstance(value, (int, float, bool)) and value < 0)):

        if not isinstance(value, int) and isinstance(min, int):
            min = vtypes.Int(min, signed=True)

        cond = value < min
        sat_val = _saturate_both(value, cond,
                                 width=width, max=max, min=min, signed=signed)
    else:
        sat_val = _saturate_overflow(value,
                                     width=width, max=max, signed=signed)

    try:
        sat_val.saturated = True
    except:
        pass

    return sat_val


def write(dst, value, blk=False, ldelay=None, rdelay=None, method='write'):
    width = vtypes.get_width(dst)
    signed = vtypes.get_signed(dst)

    if signed:
        max = _pow2(width - 1) - 1
        min = _pow2(width - 1) * (-1)
    else:
        max = _pow2(width) - 1
        min = 0

    sat_val = to_saturate(value, width=width, max=max, min=min, signed=signed)

    return getattr(dst, method)(sat_val, blk=blk, ldelay=ldelay, rdelay=rdelay)


def _add(raw_val, left, right, width=8, max=None, min=None):
    signed = vtypes.get_signed(raw_val)
    if min is None:
        min = _min(width, signed)

    if signed:
        cond = left < - right + min
        return _saturate_both(raw_val, cond,
                              width=width, max=max, min=min, signed=signed)

    return _saturate_overflow(raw_val, width=width, max=max, signed=signed)


def _sub(raw_val, left, right, width=8, max=None, min=None):
    signed = vtypes.get_signed(raw_val)
    if min is None:
        min = _min(width, signed)

    cond = left < right + min
    if signed:
        return _saturate_both(raw_val, cond,
                              width=width, max=max, min=min, signed=signed)

    return _saturate_underflow(raw_val, cond,
                               width=width, min=min, signed=signed)


def _saturate_overflow(raw_val, width=8, max=None, signed=False):
    if max is None:
        max = _max(width, signed)

    if (isinstance(max, int) and
            not isinstance(raw_val, int) and vtypes.get_signed(raw_val)):
        max = vtypes.Int(max, signed=True)

    return vtypes.Mux(raw_val > max, max, raw_val)


def _saturate_underflow(raw_val, cond, width=8, min=None, signed=False):
    if min is None:
        min = _min(width, signed)

    return vtypes.Mux(cond, min, raw_val)


def _saturate_both(raw_val, cond, width=8, max=None, min=None, signed=False):
    if max is None:
        max = _max(width, signed)

    if min is None:
        min = _min(width, signed)

    if (isinstance(max, int) and
            not isinstance(raw_val, int) and vtypes.get_signed(raw_val)):
        max = vtypes.Int(max, signed=True)

    return vtypes.Mux(cond, min, vtypes.Mux(raw_val > max, max, raw_val))


def _pow2(p):
    if isinstance(p, int):
        return 2 ** p

    return vtypes.Int(2, signed=True) ** p


def _max(width, signed=False):
    if signed:
        return _pow2(width - 1) - 1
    return _pow2(width) - 1


def _min(width, signed=False):
    if signed:
        return _pow2(width - 1) * (-1)
    return 0



from __future__ import absolute_import
from __future__ import print_function

import veriloggen.core.vtypes as vtypes
import veriloggen.core.module as module


def mkMultiplierCore(name, lsigned=True, rsigned=True, depth=6, with_update=True):

    m = module.Module(name + '_core')

    lwidth = m.Parameter('lwidth', 32)
    rwidth = m.Parameter('rwidth', 32)
    retwidth = lwidth + rwidth

    clk = m.Input('CLK')

    if with_update:
        update = m.Input('update')

    a = m.Input('a', lwidth)
    b = m.Input('b', rwidth)
    c = m.Output('c', retwidth)

    _a = m.Reg('_a', lwidth, signed=lsigned)
    _b = m.Reg('_b', rwidth, signed=rsigned)
    _mul = m.Wire('_mul', retwidth, signed=True)
    _pipe_mul = [m.Reg('_pipe_mul%d' % i, retwidth, signed=True)
                 for i in range(depth - 1)]

    __a = _a
    __b = _b
    if not lsigned:
        __a = vtypes.SystemTask(
            'signed', vtypes.Cat(vtypes.Int(0, width=1), _a))
    if not rsigned:
        __b = vtypes.SystemTask(
            'signed', vtypes.Cat(vtypes.Int(0, width=1), _b))

    m.Assign(_mul(__a * __b))
    m.Assign(c(_pipe_mul[depth - 2]))

    body = (_a(a),
            _b(b),
            _pipe_mul[0](_mul),
            [_pipe_mul[i](_pipe_mul[i - 1]) for i in range(1, depth - 1)])

    if with_update:
        body = vtypes.If(update)(body)

    m.Always(vtypes.Posedge(clk))(body)

    return m


def mkMultiplier(name, lsigned=True, rsigned=True, depth=6,
                 with_update=True, with_enable=True):

    mult = mkMultiplierCore(name, lsigned, rsigned, depth, with_update)

    m = module.Module(name)

    lwidth = m.Parameter('lwidth', 32)
    rwidth = m.Parameter('rwidth', 32)
    retwidth = lwidth + rwidth

    clk = m.Input('CLK')
    rst = m.Input('RST')

    if with_update:
        update = m.Input('update')

    if with_enable:
        enable = m.Input('enable')
    else:
        enable = 1

    if with_enable:
        valid = m.Output('valid')

    a = m.Input('a', lwidth)
    b = m.Input('b', rwidth)
    c = m.Output('c', retwidth)

    if with_enable:
        valid_reg = [m.Reg('valid_reg%d' % i) for i in range(depth)]

        m.Assign(valid(valid_reg[depth - 1]))

        body = (valid_reg[0](enable),
                [valid_reg[i](valid_reg[i - 1]) for i in range(1, depth)])

        if with_update:
            body = vtypes.If(update)(body)

        m.Always(vtypes.Posedge(clk))(
            vtypes.If(rst)(
                [valid_reg[i](0) for i in range(depth)]
            ).Else(
                body
            ))

    params = [('lwidth', lwidth), ('rwidth', rwidth)]

    ports = [('CLK', clk)]
    if with_update:
        ports.append(('update', update))

    ports.extend([('a', a), ('b', b), ('c', c)])

    m.Instance(mult, 'mult', params=params, ports=ports)

    return m


class Multiplier(object):

    def __init__(self, m, name, clk, rst, left, right, enable=None, update=None,
                 lwidth=None, rwidth=None, lsigned=True, rsigned=True, depth=6):

        if lwidth is None:
            lwidth = vtypes.get_width(left)
        if lwidth is None:
            lwidth = 1

        if rwidth is None:
            rwidth = vtypes.get_width(right)
        if rwidth is None:
            rwidth = 1

        retwidth = lwidth + rwidth

        with_enable = True if enable is not None else False
        with_update = True if update is not None else False

        self.m = m
        self.name = name
        self.clk = clk
        self.rst = rst

        self.enable = enable
        self.update = update

        if with_enable:
            self.valid = self.m.Wire(name + '_valid')

        self.left = left
        self.right = right
        self.value = self.m.Wire(name + '_value', retwidth)

        mul_def = mkMultiplier(name, lsigned, rsigned, depth,
                               with_update, with_enable)

        params = [('lwidth', lwidth), ('rwidth', rwidth)]

        ports = []
        ports.append(self.clk)
        ports.append(self.rst)

        if with_update:
            ports.append(self.update)

        if with_enable:
            ports.append(self.enable)
            ports.append(self.valid)

        ports.append(self.left)
        ports.append(self.right)
        ports.append(self.value)

        self.m.Instance(mul_def, name, params=params, ports=ports)



from __future__ import absolute_import
from __future__ import print_function
import sys
import os
import math

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(
    os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))))))

from veriloggen import *
import veriloggen.types.ram as ram


def mkMain(n=128, datawidth=32, numports=2):
    m = Module('main')

    clk = m.Input('CLK')
    rst = m.Input('RST')

    addrwidth = int(math.log(n, 2)) * 2

    myram = ram.AsyncRAM(m, 'myram', clk, datawidth, addrwidth, 2)

    seq = Seq(m, 'seq', clk, rst)

    # write
    waddr = m.Reg('waddr', 32, initval=0)
    wdata = m.Reg('wdata', 32, initval=0)
    wenable = waddr < 16

    myram.connect(0, waddr, wdata, wenable)

    seq.If(wenable)(
        waddr.inc(),
        wdata.inc()
    )

    all_ok = m.Reg('all_ok', initval=1)
    i = m.Reg('i', 32, initval=0)

    seq.If(waddr == 15, wenable == 1)(
        all_ok(1),
        i(0)
    )

    # read
    raddr = m.Reg('raddr', 32, initval=0)
    renable = Ands(raddr < 16, waddr == 16)

    myram.connect(1, raddr, 0, 0)

    seq.If(renable)(
        raddr.inc(),
    )

    rdata = myram.rdata(1)
    rvalid = renable

    sum = m.Reg('sum', 32, initval=0)

    seq.If(rvalid)(
        sum(sum + rdata),
        If(NotEql(rdata, i))(all_ok(0)),
        i.inc()
    )

    seq.Then().Delay(1)(
        Systask('display', "sum=%d", sum)
    )

    seq.If(i == 15, rvalid).Delay(1)(
        If(all_ok)(
            Display('# verify: PASSED')
        ).Else(
            Display('# verify: FAILED')
        )
    )

    seq.make_always()

    return m


def mkTest(memimg_name=None):
    m = Module('test')

    # target instance
    main = mkMain()

    # copy paras and ports
    params = m.copy_params(main)
    ports = m.copy_sim_ports(main)

    clk = ports['CLK']
    rst = ports['RST']

    uut = m.Instance(main, 'uut',
                     params=m.connect_params(main),
                     ports=m.connect_ports(main))

    # simulation.setup_waveform(m, uut, m.get_vars())
    simulation.setup_clock(m, clk, hperiod=5)
    init = simulation.setup_reset(m, rst, m.make_reset(), period=100)

    init.add(
        Delay(1000 * 100),
        Systask('finish'),
    )

    return m


def run(filename='tmp.v', simtype='iverilog', outputfile=None):

    if outputfile is None:
        outputfile = os.path.splitext(os.path.basename(__file__))[0] + '.out'

    memimg_name = 'memimg_' + outputfile

    test = mkTest(memimg_name=memimg_name)

    if filename is not None:
        test.to_verilog(filename)

    sim = simulation.Simulator(test, sim=simtype)
    rslt = sim.run(outputfile=outputfile)
    lines = rslt.splitlines()
    if simtype == 'verilator' and lines[-1].startswith('-'):
        rslt = '\n'.join(lines[:-1])
    return rslt


if __name__ == '__main__':
    rslt = run(filename='tmp.v')
    print(rslt)



from __future__ import absolute_import
from __future__ import print_function
import veriloggen
import like

expected_verilog = """
module blinkled #
 (
  parameter WIDTH = 8,
  parameter [32-1:0] WIDTH1 = 8
 )
 (
  input CLK, 
  input RST, 
  output reg [WIDTH-1:0] LED,
  output reg [32-1:0] count,
  output [32-1:0] input_count,
  output [32-1:0] output_count
 );

  localparam [32-1:0] WIDTH2 = 8;
  localparam [32-1:0] _tmp_0 = 8;

  wire [32-1:0] wire_count;
  reg [32-1:0] reg_count;
  wire [32-1:0] _tmp_1;
  reg [32-1:0] _tmp_2;

  assign wire_count = input_count;
  assign output_count = reg_count;

  always @(posedge CLK) begin
    if(RST) begin
      reg_count <= 8;
    end else begin
      reg_count <= wire_count;
    end
  end

  always @(posedge CLK) begin
    if(RST) begin        
      count <= 0;
    end else begin
      if(count == 1023) begin
        count <= 0;
      end else begin
        count <= count + 1;
      end
    end 
  end 
  always @(posedge CLK) begin
    if(RST) begin        
      LED <= 0;
    end else begin
      if(count == 1023) begin        
        LED <= LED + 1;
      end  
    end 
  end 
endmodule
"""


def test():
    veriloggen.reset()
    test_module = like.mkLed()
    code = test_module.to_verilog()

    from pyverilog.vparser.parser import VerilogParser
    from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
    parser = VerilogParser()
    expected_ast = parser.parse(expected_verilog)
    codegen = ASTCodeGenerator()
    expected_code = codegen.visit(expected_ast)

    assert(expected_code == code)



from __future__ import absolute_import
from __future__ import print_function
import veriloggen
import fsm_as_module

expected_verilog = """
module blinkled #
(
  parameter WIDTH = 8
)
(
  input CLK,
  input RST,
  output reg [WIDTH-1:0] LED
);

  reg [32-1:0] count;
  reg [32-1:0] fsm;
  localparam fsm_init = 0;
  localparam fsm_1 = 1;
  localparam fsm_2 = 2;
  localparam fsm_3 = 3;
  localparam fsm_4 = 4;
  wire [32-1:0] _fsm_out_0;

  always @(*) begin
    fsm = _fsm_out_0;
  end

  wire [32-1:0] _fsm_count_1;

  always @(*) begin
    count = _fsm_count_1;
  end

  wire [WIDTH-1:0] _fsm_LED_2;

  always @(*) begin
    LED = _fsm_LED_2;
  end

  localparam _fsm_WIDTH = WIDTH;

  sub_fsm
  #(
    .WIDTH(_fsm_WIDTH)
  )
  inst_sub_fsm
  (
    .CLK(CLK),
    .RST(RST),
    .fsm(_fsm_out_0),
    .i_count(count),
    .i_LED(LED),
    .count(_fsm_count_1),
    .LED(_fsm_LED_2)
  );


endmodule



module sub_fsm #
(
  parameter WIDTH = 8
)
(
  input CLK,
  input RST,
  output reg [32-1:0] fsm,
  input [32-1:0] i_count,
  input [WIDTH-1:0] i_LED,
  output reg [32-1:0] count,
  output reg [WIDTH-1:0] LED
);

  localparam fsm_init = 0;
  localparam fsm_1 = 1;
  localparam fsm_2 = 2;
  localparam fsm_3 = 3;
  localparam fsm_4 = 4;

  always @(posedge CLK) begin
    if(RST) begin
      fsm <= fsm_init;
      count <= 0;
      LED <= 0;
    end else begin
      case(fsm)
        fsm_init: begin
          count <= i_count + 1;
          fsm <= fsm_1;
        end
        fsm_1: begin
          count <= i_count + 2;
          fsm <= fsm_2;
        end
        fsm_2: begin
          count <= i_count + 3;
          fsm <= fsm_3;
        end
        fsm_3: begin
          if(i_count < 1024) begin
            fsm <= fsm_init;
          end else begin
            fsm <= fsm_4;
          end
        end
        fsm_4: begin
          LED <= i_LED + 1;
          fsm <= fsm_2;
        end
      endcase
    end
  end


endmodule
"""

def test():
    veriloggen.reset()
    test_module = fsm_as_module.mkLed()
    code = test_module.to_verilog()

    from pyverilog.vparser.parser import VerilogParser
    from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
    parser = VerilogParser()
    expected_ast = parser.parse(expected_verilog)
    codegen = ASTCodeGenerator()
    expected_code = codegen.visit(expected_ast)

    assert(expected_code == code)



from __future__ import absolute_import
from __future__ import print_function

import xml.dom.minidom
import datetime
import copy
import math

import veriloggen.core.vtypes as vtypes
from veriloggen.resolver import resolver
from . import axi


PORTLIST = ('AWID', 'AWADDR', 'AWLEN', 'AWSIZE', 'AWBURST', 'AWLOCK',
            'AWCACHE', 'AWPROT', 'AWQOS', 'AWUSER', 'AWVALID', 'AWREADY',
            'WDATA', 'WSTRB', 'WLAST', 'WUSER', 'WVALID', 'WREADY',
            'BID', 'BRESP', 'BUSER', 'BVALID', 'BREADY',
            'ARID', 'ARADDR', 'ARLEN', 'ARSIZE', 'ARBURST', 'ARLOCK',
            'ARCACHE', 'ARPROT', 'ARQOS', 'ARUSER', 'ARVALID', 'ARREADY',
            'RID', 'RDATA', 'RRESP', 'RLAST', 'RUSER', 'RVALID', 'RREADY')

PORTLITELIST = ('AWADDR', 'AWCACHE', 'AWPROT', 'AWVALID', 'AWREADY',
                'WDATA', 'WSTRB', 'WVALID', 'WREADY',
                'BRESP', 'BVALID', 'BREADY',
                'ARADDR', 'ARCACHE', 'ARPROT', 'ARVALID', 'ARREADY',
                'RDATA', 'RRESP', 'RVALID', 'RREADY')

PORTSTREAMLIST = ('TDATA', 'TVALID', 'TREADY', 'TLAST', 'TSTRB', 'TUSER', 'TID', 'TDEST')


class ComponentGen(object):
    def __init__(self):
        self.m = None
        self.ip_name = None
        self.bus_interfaces = None
        self.clk_ports = None
        self.rst_ports = None
        self.irq_ports = None
        self.ext_ports = None
        self.ext_params = None

        self.vendor = 'user.org'
        self.library = 'user'
        self.version = '1.0'

        self.doc = None
        self.top = None

        self.dependency_consumer = set()

    def generate(self, m, ip_name, bus_interfaces,
                 clk_ports, rst_ports, irq_ports,
                 ext_ports, ext_params,
                 vendor='user.org', library='user', version='1.0',
                 description='user description',
                 supported_families=('zynq', 'zynquplus')):

        self.m = m
        self.ip_name = ip_name
        self.bus_interfaces = bus_interfaces
        self.clk_ports = clk_ports
        self.rst_ports = rst_ports
        self.irq_ports = irq_ports
        self.ext_ports = ext_ports
        self.ext_params = ext_params

        self.resolved_m = resolver.resolve(copy.deepcopy(m))

        self.vendor = vendor
        self.library = library
        self.version = version

        if not supported_families:
            raise ValueError(
                'supported_families must be a list or tuple with valid names.')

        self.supported_families = supported_families

        impl = xml.dom.minidom.getDOMImplementation()
        self.doc = impl.createDocument('spirit', 'spirit:component', None)
        self.top = self.doc.documentElement

        self.setAttribute(self.top, 'xmlns:xilinx', "http://www.xilinx.com")
        self.setAttribute(self.top, 'xmlns:spirit',
                          "http://www.spiritconsortium.org/XMLSchema/SPIRIT/1685-2009")
        self.setAttribute(self.top, 'xmlns:xsi',
                          "http://www.w3.org/2001/XMLSchema-instance")

        self.dependency_consumer = set()

        self.top.appendChild(self.mkVendor())
        self.top.appendChild(self.mkLibrary())
        self.top.appendChild(self.mkName(self.ip_name))
        self.top.appendChild(self.mkVersion())

        self.top.appendChild(self.mkBusInterfaces())

        r = self.mkAddressSpaces()
        if r:
            self.top.appendChild(r)

        r = self.mkMemoryMaps()
        if r:
            self.top.appendChild(r)

        self.top.appendChild(self.mkModel())
        self.top.appendChild(self.mkFileSets())
        self.top.appendChild(self.mkDescription(description))
        self.top.appendChild(self.mkParameters())
        self.top.appendChild(self.mkVendorExtensions())

        return self.doc.toprettyxml(indent='  ')

    def setAttribute(self, obj, name, text):
        obj.setAttribute(name, str(text))

    def setText(self, obj, text):
        textobj = self.doc.createTextNode(str(text))
        obj.appendChild(textobj)

    def mkVendor(self):
        vendor = self.doc.createElement('spirit:vendor')
        self.setText(vendor, self.vendor)
        return vendor

    def mkLibrary(self):
        library = self.doc.createElement('spirit:library')
        self.setText(library, self.library)
        return library

    def mkVersion(self):
        version = self.doc.createElement('spirit:version')
        self.setText(version, str(self.version))
        return version

    def mkName(self, v):
        name = self.doc.createElement('spirit:name')
        self.setText(name, v)
        return name

    def mkTextNode(self, n, v):
        name = self.doc.createElement(n)
        self.setText(name, v)
        return name

    def mkBusInterfaces(self):
        bus = self.doc.createElement('spirit:busInterfaces')

        for bus_interface in self.bus_interfaces:
            bus.appendChild(self.mkBusInterface(bus_interface))

        for rst_name, rst_polarity in self.rst_ports.items():
            bus.appendChild(self.mkBusInterfaceReset(rst_name, rst_polarity))

        for clk_name, assoc_rsts in self.clk_ports.items():
            bus.appendChild(self.mkBusInterfaceClock(clk_name, assoc_rsts))

        for irq_name, sensitivity in self.irq_ports.items():
            bus.appendChild(self.mkBusInterfaceInterrupt(irq_name, sensitivity))

        return bus

    def mkBusInterface(self, obj):
        name = obj.name
        datawidth = obj.datawidth

        interface = self.doc.createElement('spirit:busInterface')
        interface.appendChild(self.mkName(name))

        if is_master(obj):
            interface.appendChild(self.mkBusTypeMemory())
            interface.appendChild(self.mkAbstractionTypeMemory())
            interface.appendChild(self.mkMaster(name))
            interface.appendChild(self.mkPortMapsMemory(obj))
        elif is_slave(obj):
            interface.appendChild(self.mkBusTypeMemory())
            interface.appendChild(self.mkAbstractionTypeMemory())
            interface.appendChild(self.mkSlave(name))
            interface.appendChild(self.mkPortMapsMemory(obj))
        elif is_streamin(obj):
            interface.appendChild(self.mkBusTypeStream())
            interface.appendChild(self.mkAbstractionTypeStream())
            interface.appendChild(self.mkStreamIn(name))
            interface.appendChild(self.mkPortMapsStream(obj))
        elif is_streamout(obj):
            interface.appendChild(self.mkBusTypeStream())
            interface.appendChild(self.mkAbstractionTypeStream())
            interface.appendChild(self.mkStreamOut(name))
            interface.appendChild(self.mkPortMapsStream(obj))
        else:
            raise TypeError("Unsupported type: '%s'" % str(type(obj)))

        return interface

    def mkBusTypeMemory(self):
        bustype = self.doc.createElement('spirit:busType')
        self.setAttribute(bustype, 'spirit:vendor', "xilinx.com")
        self.setAttribute(bustype, 'spirit:library', "interface")
        self.setAttribute(bustype, 'spirit:name', "aximm")
        self.setAttribute(bustype, 'spirit:version', "1.0")
        return bustype

    def mkAbstractionTypeMemory(self):
        abstractiontype = self.doc.createElement('spirit:abstractionType')
        self.setAttribute(abstractiontype, 'spirit:vendor', "xilinx.com")
        self.setAttribute(abstractiontype, 'spirit:library', "interface")
        self.setAttribute(abstractiontype, 'spirit:name', "aximm_rtl")
        self.setAttribute(abstractiontype, 'spirit:version', "1.0")
        return abstractiontype

    def mkBusTypeStream(self):
        bustype = self.doc.createElement('spirit:busType')
        self.setAttribute(bustype, 'spirit:vendor', "xilinx.com")
        self.setAttribute(bustype, 'spirit:library', "interface")
        self.setAttribute(bustype, 'spirit:name', "axis")
        self.setAttribute(bustype, 'spirit:version', "1.0")
        return bustype

    def mkAbstractionTypeStream(self):
        abstractiontype = self.doc.createElement('spirit:abstractionType')
        self.setAttribute(abstractiontype, 'spirit:vendor', "xilinx.com")
        self.setAttribute(abstractiontype, 'spirit:library', "interface")
        self.setAttribute(abstractiontype, 'spirit:name', "axis_rtl")
        self.setAttribute(abstractiontype, 'spirit:version', "1.0")
        return abstractiontype

    def mkMaster(self, name):
        master = self.doc.createElement('spirit:master')
        addressspaceref = self.doc.createElement('spirit:addressSpaceRef')
        self.setAttribute(addressspaceref, 'spirit:addressSpaceRef', name)
        master.appendChild(addressspaceref)
        return master

    def mkSlave(self, name):
        slave = self.doc.createElement('spirit:slave')
        memorymapref = self.doc.createElement('spirit:memoryMapRef')
        self.setAttribute(memorymapref, 'spirit:memoryMapRef', name)
        slave.appendChild(memorymapref)
        return slave

    def mkStreamIn(self, name):
        streamin = self.doc.createElement('spirit:slave')
        return streamin

    def mkStreamOut(self, name):
        streamout = self.doc.createElement('spirit:master')
        return streamout

    def mkPortMapsMemory(self, obj):
        lite = is_lite(obj)
        portmaps = self.doc.createElement('spirit:portMaps')
        portlist = list(PORTLITELIST if lite else PORTLIST)

        if not lite and obj.waddr.awid is None:
            portlist.remove('AWID')
        if not lite and obj.waddr.awuser is None:
            portlist.remove('AWUSER')
        if not lite and obj.wdata.wuser is None:
            portlist.remove('WUSER')
        if not lite and obj.wresp.bid is None:
            portlist.remove('BID')
        if not lite and obj.wresp.buser is None:
            portlist.remove('BUSER')
        if not lite and obj.raddr.arid is None:
            portlist.remove('ARID')
        if not lite and obj.raddr.aruser is None:
            portlist.remove('ARUSER')
        if not lite and obj.rdata.rid is None:
            portlist.remove('RID')
        if not lite and obj.rdata.ruser is None:
            portlist.remove('RUSER')

        for port in portlist:
            portmaps.appendChild(self.mkPortMapMemory(obj, port))

        return portmaps

    def mkPortMapsStream(self, obj):
        portmaps = self.doc.createElement('spirit:portMaps')
        portlist = list(PORTSTREAMLIST)

        if obj.tdata.tlast is None:
            portlist.remove('TLAST')
        if obj.tdata.tstrb is None:
            portlist.remove('TSTRB')
        if obj.tdata.tuser is None:
            portlist.remove('TUSER')
        if obj.tdata.tid is None:
            portlist.remove('TID')
        if obj.tdata.tdest is None:
            portlist.remove('TDEST')

        for port in portlist:
            portmaps.appendChild(self.mkPortMapStream(obj, port))

        return portmaps

    def mkPortMapMemory(self, obj, attr):
        portmap = self.doc.createElement('spirit:portMap')
        portmap.appendChild(self.mkLogicalPort(attr))
        portmap.appendChild(self.mkPhysicalPortMemory(obj, attr))
        return portmap

    def mkPortMapStream(self, obj, attr):
        portmap = self.doc.createElement('spirit:portMap')
        portmap.appendChild(self.mkLogicalPort(attr))
        portmap.appendChild(self.mkPhysicalPortStream(obj, attr))
        return portmap

    def mkLogicalPort(self, attr):
        logicalport = self.doc.createElement('spirit:logicalPort')
        logicalport.appendChild(self.mkName(attr))
        return logicalport

    def mkPhysicalPortMemory(self, obj, attr):
        if hasattr(obj.waddr, attr.lower()):
            name = getattr(obj.waddr, attr.lower()).name
        elif hasattr(obj.wdata, attr.lower()):
            name = getattr(obj.wdata, attr.lower()).name
        elif hasattr(obj.wresp, attr.lower()):
            name = getattr(obj.wresp, attr.lower()).name
        elif hasattr(obj.raddr, attr.lower()):
            name = getattr(obj.raddr, attr.lower()).name
        elif hasattr(obj.rdata, attr.lower()):
            name = getattr(obj.rdata, attr.lower()).name
        else:
            raise NameError("No such attribute '%s' in object '%s'" %
                            (attr.lower(), obj))

        physicalport = self.doc.createElement('spirit:physicalPort')
        physicalport.appendChild(self.mkName(name))
        return physicalport

    def mkPhysicalPortStream(self, obj, attr):
        if hasattr(obj.tdata, attr.lower()):
            name = getattr(obj.tdata, attr.lower()).name
        else:
            raise NameError("No such attribute '%s' in object '%s'" %
                            (attr.lower(), obj))

        physicalport = self.doc.createElement('spirit:physicalPort')
        physicalport.appendChild(self.mkName(name))
        return physicalport

    def mkBusInterfaceClock(self, name, rsts):
        interface = self.doc.createElement('spirit:busInterface')
        interface.appendChild(self.mkName(name))
        interface.appendChild(self.mkBusTypeClock())
        interface.appendChild(self.mkAbstractionTypeClock())
        interface.appendChild(self.mkSlaveClock())
        interface.appendChild(self.mkPortMapsClock(name))
        interface.appendChild(self.mkBusParametersClock(name, rsts))
        return interface

    def mkBusTypeClock(self):
        bustype = self.doc.createElement('spirit:busType')
        self.setAttribute(bustype, 'spirit:vendor', "xilinx.com")
        self.setAttribute(bustype, 'spirit:library', "signal")
        self.setAttribute(bustype, 'spirit:name', "clock")
        self.setAttribute(bustype, 'spirit:version', "1.0")
        return bustype

    def mkAbstractionTypeClock(self):
        abstractiontype = self.doc.createElement('spirit:abstractionType')
        self.setAttribute(abstractiontype, 'spirit:vendor', "xilinx.com")
        self.setAttribute(abstractiontype, 'spirit:library', "signal")
        self.setAttribute(abstractiontype, 'spirit:name', "clock_rtl")
        self.setAttribute(abstractiontype, 'spirit:version', "1.0")
        return abstractiontype

    def mkSlaveClock(self):
        slave = self.doc.createElement('spirit:slave')
        return slave

    def mkPortMapsClock(self, name):
        portmaps = self.doc.createElement('spirit:portMaps')
        portmaps.appendChild(self.mkPortMapClock(name))
        return portmaps

    def mkPortMapClock(self, name):
        portmap = self.doc.createElement('spirit:portMap')
        portmap.appendChild(self.mkLogicalPort('CLK'))
        portmap.appendChild(self.mkPhysicalPortClock(name))
        return portmap

    def mkPhysicalPortClock(self, name):
        physicalport = self.doc.createElement('spirit:physicalPort')
        physicalport.appendChild(self.mkName(name))
        return physicalport

    def mkBusParametersClock(self, name, rsts):
        parameters = self.doc.createElement('spirit:parameters')
        parameters.appendChild(self.mkBusParameterAssocBusIf(name))
        parameters.appendChild(self.mkBusParameterAssocReset(name, rsts))
        return parameters

    def mkBusParameterAssocBusIf(self, name):
        parameter = self.doc.createElement('spirit:parameter')
        parameter.appendChild(self.mkName('ASSOCIATED_BUSIF'))
        value = self.doc.createElement('spirit:value')

        bus_name_list = []
        for bus_interface in self.bus_interfaces:
            if (isinstance(bus_interface.clk, vtypes._Variable) and
                bus_interface.clk.module.is_input(bus_interface.clk.name) and
                    bus_interface.clk.name == name):
                bus_name_list.append(bus_interface.name)

        bus_names = ':'.join(bus_name_list)

        self.setAttribute(value, 'spirit:id', "BUSIFPARAM_VALUE."
                          + name + ".ASSOCIATED_BUSIF")
        self.setText(value, bus_names)
        parameter.appendChild(value)
        return parameter

    def mkBusParameterAssocReset(self, name, rsts):
        parameter = self.doc.createElement('spirit:parameter')
        parameter.appendChild(self.mkName('ASSOCIATED_RESET'))
        value = self.doc.createElement('spirit:value')

        rst_names = ':'.join(rsts)

        self.setAttribute(value, 'spirit:id', "BUSIFPARAM_VALUE."
                          + name + ".ASSOCIATED_RESET")
        self.setText(value, rst_names)
        parameter.appendChild(value)
        return parameter

    def mkBusInterfaceReset(self, name, polarity):
        interface = self.doc.createElement('spirit:busInterface')
        interface.appendChild(self.mkName(name))
        interface.appendChild(self.mkBusTypeReset())
        interface.appendChild(self.mkAbstractionTypeReset())
        interface.appendChild(self.mkSlaveReset())
        interface.appendChild(self.mkPortMapsReset(name))
        interface.appendChild(self.mkBusParametersReset(name, polarity))
        return interface

    def mkBusTypeReset(self):
        bustype = self.doc.createElement('spirit:busType')
        self.setAttribute(bustype, 'spirit:vendor', "xilinx.com")
        self.setAttribute(bustype, 'spirit:library', "signal")
        self.setAttribute(bustype, 'spirit:name', "reset")
        self.setAttribute(bustype, 'spirit:version', "1.0")
        return bustype

    def mkAbstractionTypeReset(self):
        abstractiontype = self.doc.createElement('spirit:abstractionType')
        self.setAttribute(abstractiontype, 'spirit:vendor', "xilinx.com")
        self.setAttribute(abstractiontype, 'spirit:library', "signal")
        self.setAttribute(abstractiontype, 'spirit:name', "reset_rtl")
        self.setAttribute(abstractiontype, 'spirit:version', "1.0")
        return abstractiontype

    def mkSlaveReset(self):
        slave = self.doc.createElement('spirit:slave')
        return slave

    def mkPortMapsReset(self, name):
        portmaps = self.doc.createElement('spirit:portMaps')
        portmaps.appendChild(self.mkPortMapReset(name))
        return portmaps

    def mkPortMapReset(self, name):
        portmap = self.doc.createElement('spirit:portMap')
        portmap.appendChild(self.mkLogicalPort('RST'))
        portmap.appendChild(self.mkPhysicalPortReset(name))
        return portmap

    def mkPhysicalPortReset(self, name):
        physicalport = self.doc.createElement('spirit:physicalPort')
        physicalport.appendChild(self.mkName(name))
        return physicalport

    def mkBusParametersReset(self, name, polarity):
        parameters = self.doc.createElement('spirit:parameters')
        parameters.appendChild(self.mkBusParameterPolarity(name, polarity))
        return parameters

    def mkBusParameterPolarity(self, name, polarity):
        parameter = self.doc.createElement('spirit:parameter')
        parameter.appendChild(self.mkName('POLARITY'))
        value = self.doc.createElement('spirit:value')
        self.setAttribute(value, 'spirit:id', "BUSIFPARAM_VALUE."
                          + name + ".POLARITY")
        self.setText(value, polarity)
        parameter.appendChild(value)
        return parameter

    def mkBusInterfaceInterrupt(self, name, sensitivity):
        interface = self.doc.createElement('spirit:busInterface')
        interface.appendChild(self.mkName(name))
        interface.appendChild(self.mkBusTypeInterrupt())
        interface.appendChild(self.mkAbstractionTypeInterrupt())
        interface.appendChild(self.mkMasterInterrupt())
        interface.appendChild(self.mkPortMapsInterrupt(name))
        interface.appendChild(self.mkBusParametersInterrupt(name, sensitivity))
        return interface

    def mkBusTypeInterrupt(self):
        bustype = self.doc.createElement('spirit:busType')
        self.setAttribute(bustype, 'spirit:vendor', "xilinx.com")
        self.setAttribute(bustype, 'spirit:library', "signal")
        self.setAttribute(bustype, 'spirit:name', "interrupt")
        self.setAttribute(bustype, 'spirit:version', "1.0")
        return bustype

    def mkAbstractionTypeInterrupt(self):
        abstractiontype = self.doc.createElement('spirit:abstractionType')
        self.setAttribute(abstractiontype, 'spirit:vendor', "xilinx.com")
        self.setAttribute(abstractiontype, 'spirit:library', "signal")
        self.setAttribute(abstractiontype, 'spirit:name', "interrupt_rtl")
        self.setAttribute(abstractiontype, 'spirit:version', "1.0")
        return abstractiontype

    def mkMasterInterrupt(self):
        master = self.doc.createElement('spirit:master')
        return master

    def mkPortMapsInterrupt(self, name):
        portmaps = self.doc.createElement('spirit:portMaps')
        portmaps.appendChild(self.mkPortMapInterrupt(name))
        return portmaps

    def mkPortMapInterrupt(self, name):
        portmap = self.doc.createElement('spirit:portMap')
        portmap.appendChild(self.mkLogicalPort('INTERRUPT'))
        portmap.appendChild(self.mkPhysicalPortInterrupt(name))
        return portmap

    def mkPhysicalPortInterrupt(self, name):
        physicalport = self.doc.createElement('spirit:physicalPort')
        physicalport.appendChild(self.mkName(name))
        return physicalport

    def mkBusParametersInterrupt(self, name, sensitivity):
        parameters = self.doc.createElement('spirit:parameters')
        parameters.appendChild(self.mkBusParameterSensitivity(name, sensitivity))
        return parameters

    def mkBusParameterSensitivity(self, name, sensitivity):
        parameter = self.doc.createElement('spirit:parameter')
        parameter.appendChild(self.mkName('SENSITIVITY'))
        value = self.doc.createElement('spirit:value')
        self.setAttribute(value, 'spirit:id', "BUSIFPARAM_VALUE."
                          + name.upper() + ".SENSITIVITY")
        self.setText(value, sensitivity)
        parameter.appendChild(value)
        return parameter

    def mkAddressSpaces(self):
        isempty = True
        spaces = self.doc.createElement('spirit:addressSpaces')
        for bus_interface in self.bus_interfaces:
            if isinstance(bus_interface, (axi.AxiLiteMaster, axi.AxiMaster)):
                spaces.appendChild(self.mkAddressSpace(bus_interface))
                isempty = False
        if isempty:
            return None
        return spaces

    def mkAddressSpace(self, obj):
        name = obj.name

        space = self.doc.createElement('spirit:addressSpace')
        space.appendChild(self.mkName(name))

        range = self.doc.createElement('spirit:range')
        self.setAttribute(range, 'spirit:format', "long")
        range_value = 2 ** self.resolved_m[obj.name + '_awaddr'].width
        self.setText(range, range_value)
        space.appendChild(range)

        width = self.doc.createElement('spirit:width')
        self.setAttribute(width, 'spirit:format', "long")
        width_value = self.resolved_m[obj.name + '_wdata'].width
        self.setText(width, width_value)
        space.appendChild(width)

        return space

    def mkMemoryMaps(self):
        isempty = True
        maps = self.doc.createElement('spirit:memoryMaps')
        for bus_interface in self.bus_interfaces:
            if isinstance(bus_interface, (axi.AxiLiteSlave, axi.AxiSlave)):
                maps.appendChild(self.mkMemoryMap(bus_interface))
                isempty = False
        if isempty:
            return None
        return maps

    def mkMemoryMap(self, obj):
        name = obj.name
        map = self.doc.createElement('spirit:memoryMap')
        map.appendChild(self.mkName(name))

        addressblock = self.doc.createElement('spirit:addressBlock')
        addressblock.appendChild(self.mkName(name + '_reg'))

        baseaddr = self.doc.createElement('spirit:baseAddress')
        self.setAttribute(baseaddr, 'spirit:format', "long")
        self.setText(baseaddr, 0)
        addressblock.appendChild(baseaddr)

        range = self.doc.createElement('spirit:range')
        self.setAttribute(range, 'spirit:format', "long")

        if hasattr(obj, 'register') and isinstance(obj.register, (tuple, list)):
            map_range = 2 ** int(math.ceil(math.log(max(len(obj.register), 4096), 2)))
        else:
            map_range = 2 ** self.resolved_m[obj.name + '_awaddr'].width

        self.setText(range, map_range)
        addressblock.appendChild(range)

        width = self.doc.createElement('spirit:width')
        self.setAttribute(width, 'spirit:format', "long")
        self.setText(width, obj.datawidth)
        addressblock.appendChild(width)

        usage = self.doc.createElement('spirit:usage')
        self.setText(usage, 'register')
        addressblock.appendChild(usage)

        map.appendChild(addressblock)

        return map

    def mkModel(self):
        model = self.doc.createElement('spirit:model')
        model.appendChild(self.mkViews())
        model.appendChild(self.mkPorts())
        return model

    def mkViews(self):
        views = self.doc.createElement('spirit:views')
        views.appendChild(self.mkView('xilinx_verilogsynthesis',
                                      'Verilog Synthesis',
                                      'verilogSource:vivado.xilinx.com:synthesis',
                                      'verilog',
                                      self.ip_name,
                                      'xilinx_verilogsynthesis_view_fileset'))
        views.appendChild(self.mkView('xilinx_verilogbehavioralsimulation',
                                      'Verilog Simulation',
                                      'verilogSource:vivado.xilinx.com:simulation',
                                      'verilog',
                                      self.ip_name,
                                      'xilinx_verilogbehavioralsimulation_view_fileset'))
        views.appendChild(self.mkView('xilinx_xpgui',
                                      'UI Layout',
                                      ':vivado.xilinx.com:xgui.ui',
                                      None,
                                      None,
                                      'xilinx_xpgui_view_fileset'))
        return views

    def mkView(self, name, displayname, envidentifier, language, modelname, localname):
        view = self.doc.createElement('spirit:view')
        view.appendChild(self.mkName(name))
        view.appendChild(self.mkTextNode('spirit:displayName', displayname))
        view.appendChild(self.mkTextNode('spirit:envIdentifier', envidentifier))

        if language is not None:
            view.appendChild(self.mkTextNode('spirit:language', language))

        if modelname is not None:
            view.appendChild(self.mkTextNode('spirit:modelName', modelname))

        filesetref = self.doc.createElement('spirit:fileSetRef')
        filesetref.appendChild(self.mkTextNode('spirit:localName', localname))
        view.appendChild(filesetref)

        return view

    def mkPorts(self):
        ports = self.doc.createElement('spirit:ports')

        for portname, port in self.ext_ports.items():
            ports.appendChild(self.mkPortSignal(port))

        return ports

    def mkPortSignal(self, var):
        name = var.name
        direction = ('in' if isinstance(var, vtypes.Input) else
                     'out' if isinstance(var, vtypes.Output) else
                     'inout')

        width = self.resolved_m[name].width
        h = width - 1 if width is not None else None
        l = 0 if h is not None else None

        return self.mkPortEntry(name, direction,
                                None, h, None, l)

    def mkPortEntry(self, name, direction,
                    lvar, lvalue, rvar, rvalue,
                    withdriver=False,
                    extensionvar=None, extensionvalue='true'):

        port = self.doc.createElement('spirit:port')
        port.appendChild(self.mkName(name))
        port.appendChild(self.mkWire(direction, lvar, lvalue, rvar, rvalue, withdriver))

        if extensionvar is not None:
            port.appendChild(
                self.mkPortVendorExtensions(name, extensionvar, extensionvalue))

        return port

    def mkWire(self, direction, lvar, lvalue, rvar, rvalue, withdriver=False):
        wire = self.doc.createElement('spirit:wire')
        wire.appendChild(self.mkDirection(direction))

        if not (lvalue is None and rvalue is None):
            wire.appendChild(self.mkVector(lvar, lvalue, rvar, rvalue))

        wire.appendChild(self.mkWireTypeDefs('wire'))

        if withdriver:
            wire.appendChild(self.mkDriver())

        return wire

    def mkDirection(self, direction):
        return self.mkTextNode('spirit:direction', direction)

    def mkVector(self, lvar, lvalue, rvar, rvalue):
        vector = self.doc.createElement('spirit:vector')

        left = self.doc.createElement('spirit:left')
        self.setAttribute(left, 'spirit:format', "long")

        if lvar is not None:
            self.setAttribute(left, 'spirit:resolve', 'dependent')
            self.setAttribute(left, 'spirit:dependency', lvar)

        self.setText(left, lvalue)
        vector.appendChild(left)

        right = self.doc.createElement('spirit:right')
        self.setAttribute(right, 'spirit:format', "long")

        if rvar is not None:
            self.setAttribute(right, 'spirit:resolve', 'dependent')
            self.setAttribute(right, 'spirit:dependency', rvar)

        self.setText(right, rvalue)
        vector.appendChild(right)

        return vector

    def mkWireTypeDefs(self, wiretype):
        wiretypedefs = self.doc.createElement('spirit:wireTypeDefs')
        wiretypedefs.appendChild(self.mkWireTypeDef(wiretype))
        return wiretypedefs

    def mkWireTypeDef(self, wiretype):
        wiretypedef = self.doc.createElement('spirit:wireTypeDef')
        wiretypedef.appendChild(self.mkTextNode('spirit:typeName', wiretype))
        wiretypedef.appendChild(
            self.mkTextNode('spirit:viewNameRef', 'xilinx_verilogsynthesis'))
        wiretypedef.appendChild(
            self.mkTextNode('spirit:viewNameRef', 'xilinx_verilogbehavioralsimulation'))
        return wiretypedef

    def mkDriver(self):
        driver = self.doc.createElement('spirit:driver')
        driver.appendChild(self.mkTextNode('spirit:defaultValue', 0))
        return driver

    def mkPortVendorExtensions(self, name, var, value='true'):
        extensions = self.doc.createElement('spirit:vendorExtensions')
        portinfo = self.doc.createElement('xilinx:portInfo')
        enablement = self.doc.createElement('xilinx:enablement')

        isEnabled = self.doc.createElement('xilinx:isEnabled')
        self.setAttribute(isEnabled, 'xilinx:resolve', "dependent")
        self.setAttribute(isEnabled, 'xilinx:id', 'PORT_ENABLEMENT.' + name)
        self.setAttribute(isEnabled, 'xilinx:dependency',
                          "spirit:decode(id('MODELPARAM_VALUE." + var.name + "')) >0")
        self.setText(isEnabled, value)
        enablement.appendChild(isEnabled)

        portinfo.appendChild(enablement)
        extensions.appendChild(portinfo)
        return extensions

    def mkFileSets(self):
        filesets = self.doc.createElement('spirit:fileSets')

        source = self.doc.createElement('spirit:fileSet')
        source.appendChild(self.mkName("xilinx_verilogsynthesis_view_fileset"))
        source.appendChild(self.mkFileSet('hdl/' + self.ip_name + '.v',
                                          'verilogSource'))
        filesets.appendChild(source)

        sim = self.doc.createElement('spirit:fileSet')
        sim.appendChild(self.mkName(
            "xilinx_verilogbehavioralsimulation_view_fileset"))
        sim.appendChild(self.mkFileSet('hdl/' + self.ip_name + '.v',
                                       'verilogSource'))
        filesets.appendChild(sim)

        xguitcl = self.doc.createElement('spirit:fileSet')
        xguitcl.appendChild(self.mkName("xilinx_xpgui_view_fileset"))
        xguitcl.appendChild(self.mkFileSet('xgui/xgui.tcl',
                                           'tclSource', 'XGUI_VERSION_2'))
        filesets.appendChild(xguitcl)

        return filesets

    def mkFileSet(self, name, filetype=None, *userfiletypes):
        fileset = self.doc.createElement('spirit:file')
        fileset.appendChild(self.mkName(name))
        if filetype is not None:
            fileset.appendChild(self.mkTextNode('spirit:fileType', filetype))
        for u in userfiletypes:
            fileset.appendChild(self.mkTextNode('spirit:userFileType', u))
        return fileset

    def mkDescription(self, description):
        return self.mkTextNode('spirit:description', description)

    def mkParameters(self):
        parameters = self.doc.createElement('spirit:parameters')

        compname = self.doc.createElement('spirit:parameter')
        compname.appendChild(self.mkName('Component_Name'))
        value = self.doc.createElement('spirit:value')
        self.setAttribute(value, 'spirit:resolve', 'user')
        self.setAttribute(value, 'spirit:id',
                          "PARAM_VALUE." + 'Component_Name')
        self.setAttribute(value, 'spirit:order', 1)
        self.setText(value, self.ip_name + '_v' + self.version.replace('.', '_'))
        compname.appendChild(value)
        parameters.appendChild(compname)

        return parameters

    def mkVendorExtensions(self):
        extensions = self.doc.createElement('spirit:vendorExtensions')
        extensions.appendChild(self.mkCoreExtensions())
        packageinfo = self.doc.createElement('xilinx:packagingInfo')
        packageinfo.appendChild(self.mkTextNode(
            'xilinx:xilinxVersion', '2018.3'))
        extensions.appendChild(packageinfo)
        return extensions

    def mkCoreExtensions(self):
        coreextensions = self.doc.createElement('xilinx:coreExtensions')
        supported = self.doc.createElement('xilinx:supportedFamilies')

        for name in self.supported_families:
            family = self.doc.createElement('xilinx:family')
            self.setAttribute(family, 'xilinx:lifeCycle', 'Production')
            self.setText(family, name)
            supported.appendChild(family)

        coreextensions.appendChild(supported)

        taxonomies = self.doc.createElement('xilinx:taxonomies')
        taxonomies.appendChild(self.mkTextNode('xilinx:taxonomy', '/UserIP'))
        coreextensions.appendChild(taxonomies)

        coreextensions.appendChild(
            self.mkTextNode('xilinx:displayName',
                            (self.ip_name + '_v' + self.version.replace('.', '_'))))

        coreextensions.appendChild(self.mkTextNode('xilinx:coreRevision', 1))

        now = datetime.datetime.now()
        dt = now.strftime("%Y-%m-%dT%H:%M:%SZ")  # '2015-03-08T02:16:15Z'
        coreextensions.appendChild(self.mkTextNode('xilinx:coreCreationDateTime', dt))

        return coreextensions


def is_master(obj):
    return isinstance(obj, (axi.AxiMaster, axi.AxiLiteMaster))


def is_slave(obj):
    return isinstance(obj, (axi.AxiSlave, axi.AxiLiteSlave))


def is_streamin(obj):
    return isinstance(obj, axi.AxiStreamIn)


def is_streamout(obj):
    return isinstance(obj, axi.AxiStreamOut)


def is_lite(obj):
    return isinstance(obj, (axi.AxiLiteMaster, axi.AxiLiteSlave))



from __future__ import absolute_import
from __future__ import print_function
import sys
import os

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(
    os.path.dirname(os.path.abspath(__file__)))))

from veriloggen import *
import veriloggen.thread as vthread
import veriloggen.types.axi as axi
import veriloggen.types.ipxact as ipxact


def mkMemcpy():
    m = Module('memcpy')
    clk = m.Input('CLK')
    rst = m.Input('RST')

    datawidth = 32
    addrwidth = 10

    ram_words = (2 ** addrwidth) // (datawidth // 8)

    ram_a = vthread.RAM(m, 'ram_a', clk, rst, datawidth, addrwidth)
    maxi = vthread.AXIM(m, 'maxi', clk, rst, datawidth)
    saxi = vthread.AXISLiteRegister(m, 'saxi', clk, rst, datawidth, length=8)

    def memcpy():
        while True:
            saxi.wait_flag(0, value=1, resetvalue=0)

            copy_bytes = saxi.read(1)
            src_offset = saxi.read(2)
            dst_offset = saxi.read(3)

            copy(copy_bytes, src_offset, dst_offset)

            saxi.write_flag(4, 1, resetvalue=0)

    def copy(copy_bytes, src_offset, dst_offset):
        rest_words = copy_bytes // (datawidth // 8)
        src_global_addr = src_offset
        dst_global_addr = dst_offset
        local_addr = 0

        while rest_words > 0:
            if rest_words > ram_words:
                dma_size = ram_words
            else:
                dma_size = rest_words

            maxi.dma_read(ram_a, local_addr, src_global_addr, dma_size)
            maxi.dma_write(ram_a, local_addr, dst_global_addr, dma_size)

            src_global_addr += dma_size * (datawidth // 8)
            dst_global_addr += dma_size * (datawidth // 8)
            rest_words -= dma_size

    th = vthread.Thread(m, 'th_memcpy', clk, rst, memcpy)
    fsm = th.start()

    return m


def mkTest():
    m = Module('test')

    copy_bytes = 1024 * 4

    # target instance
    memcpy = mkMemcpy()

    uut = Submodule(m, memcpy, name='uut')
    clk = uut['CLK']
    rst = uut['RST']

    memory = axi.AxiMemoryModel(m, 'memory', clk, rst)
    memory.connect(uut.get_inst_ports(), 'maxi')

    # AXI-Slave controller
    _saxi = vthread.AXIMLite(m, '_saxi', clk, rst, noio=True)
    _saxi.connect(uut.get_inst_ports(), 'saxi')

    # Timer
    counter = m.Reg('counter', 32, initval=0)
    seq = Seq(m, 'seq', clk, rst)
    seq(
        counter.inc()
    )

    def ctrl():
        for i in range(100):
            pass

        awaddr = 4 * 1
        print('# copy_bytes = %d' % copy_bytes)
        _saxi.write(awaddr, copy_bytes)

        awaddr = 4 * 2
        src_offset = 0
        print('# src_offset = %d' % src_offset)
        _saxi.write(awaddr, src_offset)

        awaddr = 4 * 3
        dst_offset = 1024 * 8
        print('# dst_offset = %d' % dst_offset)
        _saxi.write(awaddr, dst_offset)

        awaddr = 4 * 0
        start_time = counter
        print('# start time = %d' % start_time)
        _saxi.write(awaddr, 1)

        araddr = 4 * 4
        v = _saxi.read(araddr)
        while v == 0:
            v = _saxi.read(araddr)

        end_time = counter
        print('# end time = %d' % end_time)
        time = end_time - start_time
        print('# exec time = %d' % time)

    th = vthread.Thread(m, 'th_ctrl', clk, rst, ctrl)
    fsm = th.start()

    simulation.setup_waveform(m, uut)
    simulation.setup_clock(m, clk, hperiod=5)
    init = simulation.setup_reset(m, rst, m.make_reset(), period=100)

    init.add(
        Delay(1000000),
        Systask('finish'),
    )

    return m


if __name__ == '__main__':
    test = mkTest()
    verilog = test.to_verilog('tmp.v')
    print(verilog)

    sim = simulation.Simulator(test)
    rslt = sim.run()
    print(rslt)

    m = mkMemcpy()
    ipxact.to_ipxact(m,
                     clk_ports=[('CLK', ('RST',))],
                     rst_ports=[('RST', 'ACTIVE_HIGH')])



from __future__ import absolute_import
from __future__ import print_function
import veriloggen
import seq_countup_if_then

expected_verilog = """
module test #
(
  parameter INTERVAL = 16
);

  reg CLK;
  reg RST;
  wire [8-1:0] LED;

  blinkled
  #(
    .INTERVAL(INTERVAL)
  )
  uut
  (
    .CLK(CLK),
    .RST(RST),
    .LED(LED)
  );


  initial begin
    CLK = 0;
    forever begin
      #5 CLK = !CLK;
    end
  end


  initial begin
    RST = 0;
    #100;
    RST = 1;
    #100;
    RST = 0;
    #1000;
    $finish;
  end


endmodule



module blinkled #
(
  parameter INTERVAL = 16
)
(
  input CLK,
  input RST,
  output reg [8-1:0] LED
);

  reg [32-1:0] count;
  reg _tmp_0;
  reg updated;
  reg _seq_cond_0_1;
  reg _seq_cond_1_1;
  reg _seq_cond_1_2;

  always @(posedge CLK) begin
    if(RST) begin
      count <= 0;
      LED <= 0;
      _seq_cond_0_1 <= 0;
      updated <= 0;
      _seq_cond_1_1 <= 0;
      _seq_cond_1_2 <= 0;
    end else begin
      if(_seq_cond_1_2) begin
        updated <= 0;
      end 
      if(_seq_cond_0_1) begin
        updated <= 1;
      end 
      _seq_cond_1_2 <= _seq_cond_1_1;
      $display("LED:%d count:%d", LED, count);
      if(count < INTERVAL - 1) begin
        count <= count + 1;
      end else begin
        count <= 0;
        LED <= LED + 1;
      end
      _seq_cond_0_1 <= !(count < INTERVAL - 1);
      _seq_cond_1_1 <= !(count < INTERVAL - 1);
    end
  end


endmodule
"""

def test():
    veriloggen.reset()
    test_module = seq_countup_if_then.mkTest()
    code = test_module.to_verilog()

    from pyverilog.vparser.parser import VerilogParser
    from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
    parser = VerilogParser()
    expected_ast = parser.parse(expected_verilog)
    codegen = ASTCodeGenerator()
    expected_code = codegen.visit(expected_ast)

    assert(expected_code == code)



from __future__ import absolute_import
from __future__ import print_function
import veriloggen
import thread_slice

expected_verilog = """
module test;

  reg CLK;
  reg RST;
  wire [8-1:0] LED;

  blinkled
  uut
  (
    .CLK(CLK),
    .RST(RST),
    .LED(LED)
  );


  initial begin
    CLK = 0;
    forever begin
      #5 CLK = !CLK;
    end
  end


  initial begin
    RST = 0;
    #100;
    RST = 1;
    #100;
    RST = 0;
    #10000;
    $finish;
  end


endmodule



module blinkled
(
  input CLK,
  input RST,
  output reg [8-1:0] LED
);

  reg [8-1:0] count;
  reg [32-1:0] th_blink;
  localparam th_blink_init = 0;
  reg signed [32-1:0] _th_blink_times_0;
  reg signed [32-1:0] _th_blink_i_1;
  reg signed [32-1:0] _th_blink_x_2;
  localparam th_blink_1 = 1;
  localparam th_blink_2 = 2;
  localparam th_blink_3 = 3;
  localparam th_blink_4 = 4;
  localparam th_blink_5 = 5;
  localparam th_blink_6 = 6;
  localparam th_blink_7 = 7;
  localparam th_blink_8 = 8;
  localparam th_blink_9 = 9;
  localparam th_blink_10 = 10;
  localparam th_blink_11 = 11;
  localparam th_blink_12 = 12;
  localparam th_blink_13 = 13;
  localparam th_blink_14 = 14;
  localparam th_blink_15 = 15;
  localparam th_blink_16 = 16;
  localparam th_blink_17 = 17;
  localparam th_blink_18 = 18;
  localparam th_blink_19 = 19;
  localparam th_blink_20 = 20;
  localparam th_blink_21 = 21;
  localparam th_blink_22 = 22;
  localparam th_blink_23 = 23;
  localparam th_blink_24 = 24;
  localparam th_blink_25 = 25;
  localparam th_blink_26 = 26;
  localparam th_blink_27 = 27;
  localparam th_blink_28 = 28;

  always @(posedge CLK) begin
    if(RST) begin
      th_blink <= th_blink_init;
      _th_blink_times_0 <= 0;
      LED <= 0;
      count <= 0;
      _th_blink_i_1 <= 0;
      _th_blink_x_2 <= 0;
      LED[_th_blink_x_2] <= (0 >> _th_blink_x_2) & 1'd1;
    end else begin
      case(th_blink)
        th_blink_init: begin
          _th_blink_times_0 <= 10;
          th_blink <= th_blink_1;
        end
        th_blink_1: begin
          LED <= 0;
          th_blink <= th_blink_2;
        end
        th_blink_2: begin
          count <= 0;
          th_blink <= th_blink_3;
        end
        th_blink_3: begin
          _th_blink_i_1 <= 0;
          th_blink <= th_blink_4;
        end
        th_blink_4: begin
          if(_th_blink_i_1 < _th_blink_times_0) begin
            th_blink <= th_blink_5;
          end else begin
            th_blink <= th_blink_12;
          end
        end
        th_blink_5: begin
          _th_blink_x_2 <= 0;
          th_blink <= th_blink_6;
        end
        th_blink_6: begin
          if(_th_blink_x_2 < 8) begin
            th_blink <= th_blink_7;
          end else begin
            th_blink <= th_blink_9;
          end
        end
        th_blink_7: begin
          LED[_th_blink_x_2] <= count[_th_blink_x_2];
          th_blink <= th_blink_8;
        end
        th_blink_8: begin
          _th_blink_x_2 <= _th_blink_x_2 + 1;
          th_blink <= th_blink_6;
        end
        th_blink_9: begin
          $display("led =  %d", LED);
          th_blink <= th_blink_10;
        end
        th_blink_10: begin
          count <= count + 1;
          th_blink <= th_blink_11;
        end
        th_blink_11: begin
          _th_blink_i_1 <= _th_blink_i_1 + 1;
          th_blink <= th_blink_4;
        end
        th_blink_12: begin
          LED <= 0;
          th_blink <= th_blink_13;
        end
        th_blink_13: begin
          count <= 0;
          th_blink <= th_blink_14;
        end
        th_blink_14: begin
          _th_blink_i_1 <= 0;
          th_blink <= th_blink_15;
        end
        th_blink_15: begin
          if(_th_blink_i_1 < _th_blink_times_0) begin
            th_blink <= th_blink_16;
          end else begin
            th_blink <= th_blink_20;
          end
        end
        th_blink_16: begin
          LED <= count[1:0];
          th_blink <= th_blink_17;
        end
        th_blink_17: begin
          $display("led =  %d", LED);
          th_blink <= th_blink_18;
        end
        th_blink_18: begin
          count <= count + 1;
          th_blink <= th_blink_19;
        end
        th_blink_19: begin
          _th_blink_i_1 <= _th_blink_i_1 + 1;
          th_blink <= th_blink_15;
        end
        th_blink_20: begin
          LED <= 0;
          th_blink <= th_blink_21;
        end
        th_blink_21: begin
          count <= 0;
          th_blink <= th_blink_22;
        end
        th_blink_22: begin
          _th_blink_i_1 <= 0;
          th_blink <= th_blink_23;
        end
        th_blink_23: begin
          if(_th_blink_i_1 < _th_blink_times_0) begin
            th_blink <= th_blink_24;
          end else begin
            th_blink <= th_blink_28;
          end
        end
        th_blink_24: begin
          LED <= { count[6], count[4], count[2], count[0] };
          th_blink <= th_blink_25;
        end
        th_blink_25: begin
          $display("led =  %d", LED);
          th_blink <= th_blink_26;
        end
        th_blink_26: begin
          count <= count + 1;
          th_blink <= th_blink_27;
        end
        th_blink_27: begin
          _th_blink_i_1 <= _th_blink_i_1 + 1;
          th_blink <= th_blink_23;
        end
      endcase
    end
  end


endmodule
"""


def test():
    veriloggen.reset()
    test_module = thread_slice.mkTest()
    code = test_module.to_verilog()

    from pyverilog.vparser.parser import VerilogParser
    from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
    parser = VerilogParser()
    expected_ast = parser.parse(expected_verilog)
    codegen = ASTCodeGenerator()
    expected_code = codegen.visit(expected_ast)

    assert(expected_code == code)



from __future__ import absolute_import
from __future__ import print_function
import veriloggen
import types_rom_sync

expected_verilog = """
module test;

  reg CLK;
  reg RST;

  main
  uut
  (
    .CLK(CLK),
    .RST(RST)
  );


  initial begin
    $dumpfile("uut.vcd");
    $dumpvars(0, uut, CLK, RST);
  end


  initial begin
    CLK = 0;
    forever begin
      #5 CLK = !CLK;
    end
  end


  initial begin
    RST = 0;
    #100;
    RST = 1;
    #100;
    RST = 0;
    #100000;
    $finish;
  end


endmodule



module main
(
  input CLK,
  input RST
);

  reg [4-1:0] addr;
  wire [4-1:0] myrom_addr;
  assign myrom_addr = addr;
  wire [8-1:0] myrom_val;

  myrom
  myrom
  (
    .CLK(CLK),
    .addr(myrom_addr),
    .val(myrom_val)
  );


  always @(posedge CLK) begin
    if(RST) begin
      addr <= 0;
    end else begin
      addr <= addr + 1;
      $display("addr=%d rdata=%d", addr, myrom_val);
    end
  end


endmodule



module myrom
(
  input CLK,
  input [4-1:0] addr,
  output reg [8-1:0] val
);


  always @(posedge CLK) begin
    case(addr)
      0: begin
        val <= 0;
      end
      1: begin
        val <= 1;
      end
      2: begin
        val <= 4;
      end
      3: begin
        val <= 9;
      end
      4: begin
        val <= 16;
      end
      5: begin
        val <= 25;
      end
      6: begin
        val <= 36;
      end
      7: begin
        val <= 49;
      end
      8: begin
        val <= 64;
      end
      9: begin
        val <= 81;
      end
      10: begin
        val <= 100;
      end
      11: begin
        val <= 121;
      end
      12: begin
        val <= 144;
      end
      13: begin
        val <= 169;
      end
      14: begin
        val <= 196;
      end
      15: begin
        val <= 225;
      end
    endcase
  end


endmodule
"""


def test():
    veriloggen.reset()
    test_module = types_rom_sync.mkTest()
    code = test_module.to_verilog()

    from pyverilog.vparser.parser import VerilogParser
    from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
    parser = VerilogParser()
    expected_ast = parser.parse(expected_verilog)
    codegen = ASTCodeGenerator()
    expected_code = codegen.visit(expected_ast)

    assert(expected_code == code)



from __future__ import absolute_import
from __future__ import print_function

import math
import functools
from collections import OrderedDict

import veriloggen.core.vtypes as vtypes
import veriloggen.types.axi as axi
from veriloggen.fsm.fsm import FSM
from veriloggen.optimizer import try_optimize as optimize

from .ttypes import _MutexFunction
from .ram import RAM, MultibankRAM, to_multibank_ram
from .fifo import FIFO


class AXIStreamOut(axi.AxiStreamOut, _MutexFunction):
    """ AXI Stream Interface for Output """

    __intrinsics__ = ('write',
                      'read_ram', 'read_ram_async',
                      'wait_read_ram')

    def __init__(self, m, name, clk, rst, datawidth=32, addrwidth=32,
                 with_last=True, with_strb=False,
                 id_width=0, user_width=0, dest_width=0,
                 noio=False,
                 enable_async=True,
                 num_cmd_delay=0, num_data_delay=0,
                 op_sel_width=8, fsm_as_module=False):

        axi.AxiStreamOut.__init__(self, m, name, clk, rst, datawidth,
                                  with_last, with_strb,
                                  id_width, user_width, dest_width,
                                  noio)

        self.addrwidth = addrwidth

        self.enable_async = enable_async
        self.num_cmd_delay = num_cmd_delay
        self.num_data_delay = num_data_delay
        self.op_sel_width = op_sel_width
        self.fsm_as_module = fsm_as_module

        self.mutex = None

        self.write_start = self.m.Reg('_'.join(['', self.name, 'write_start']),
                                      initval=0)
        self.write_op_sel = self.m.Reg('_'.join(['', self.name, 'write_op_sel']),
                                       self.op_sel_width, initval=0)
        self.write_local_addr = self.m.Reg('_'.join(['', self.name, 'write_local_addr']),
                                           self.addrwidth, initval=0)
        self.write_size = self.m.Reg('_'.join(['', self.name, 'write_size']),
                                     self.addrwidth + 1, initval=0)
        self.write_local_stride = self.m.Reg('_'.join(['', self.name, 'write_local_stride']),
                                             self.addrwidth, initval=0)
        self.write_idle = self.m.Reg(
            '_'.join(['', self.name, 'write_idle']), initval=1)

        self.seq(
            self.write_start(0)
        )

        self.write_op_id_map = OrderedDict()
        self.write_op_id_count = 1
        self.write_reqs = OrderedDict()
        self.write_ops = []

        self.write_fsm = None
        self.write_counter = None

        self.write_narrow_fsms = OrderedDict()  # key: pack_size
        self.write_narrow_counters = OrderedDict()  # key: pack_size
        self.write_narrow_wdatas = OrderedDict()  # key: pack_size
        self.write_narrow_wvalids = OrderedDict()  # key: pack_size
        self.write_narrow_wlasts = OrderedDict()  # key: pack_size
        self.write_narrow_pack_counts = OrderedDict()  # key: pack_size

        self.write_wide_fsms = OrderedDict()  # key: pack_size
        self.write_wide_counters = OrderedDict()  # key: pack_size
        self.write_wide_wdatas = OrderedDict()  # key: pack_size
        self.write_wide_wvalids = OrderedDict()  # key: pack_size
        self.write_wide_wlasts = OrderedDict()  # key: pack_size
        self.write_wide_pack_counts = OrderedDict()  # key: pack_size

    def write(self, fsm, value, last=False):
        ack = self.write_data(value, last, cond=fsm)
        fsm.If(ack).goto_next()

    def read_ram(self, fsm, ram, local_addr, size,
                 local_stride=1, port=0, ram_method=None):

        if self.enable_async:
            self.wait_read_ram(fsm)

        self._read_ram(fsm, ram, local_addr, size,
                       local_stride, port, ram_method)

        self.wait_read_ram(fsm)

    def read_ram_async(self, fsm, ram, local_addr, size,
                       local_stride=1, port=0, ram_method=None):

        if not self.enable_async:
            raise ValueError(
                "Async mode is disabled. Set 'True' to AXIM.enable_async.")

        self.wait_read_ram(fsm)

        self._read_ram(fsm, ram, local_addr, size,
                       local_stride, port, ram_method)

    def wait_read_ram(self, fsm):

        fsm.If(self.write_idle).goto_next()

    def _read_ram(self, fsm, ram, local_addr, size,
                  local_stride=1, port=0, ram_method=None):

        if isinstance(ram, (tuple, list)):
            ram = to_multibank_ram(ram)

        if not isinstance(ram, (RAM, MultibankRAM)):
            raise TypeError('RAM object is required.')

        if ram_method is None:
            ram_method = getattr(ram, 'read_dataflow')

        start = self._set_flag(fsm)

        for _ in range(self.num_cmd_delay + 1):
            fsm.goto_next()

        self._set_write_request(ram, port, ram_method, start,
                                local_addr, size, local_stride)

        self._synthesize_write_fsm(ram, port, ram_method)

        fsm.goto_next()

    def _set_write_request(self, ram, port, ram_method, start,
                           local_addr, size, local_stride):

        op_id = self._get_write_op_id(ram, port, ram_method)

        if op_id in self.write_reqs:
            (write_start, write_op_sel,
             write_local_addr_in,
             write_size_in, write_local_stride_in) = self.write_reqs[op_id]

            self.seq.If(start)(
                write_start(1),
                write_op_sel(op_id),
                write_local_addr_in(local_addr),
                write_size_in(size),
                write_local_stride_in(local_stride)
            )

            return

        port = str(vtypes.to_int(port))

        write_start = self.m.Reg(
            '_'.join(['', self.name, ram.name, port, 'write_start']),
            initval=0)
        write_op_sel = self.m.Reg(
            '_'.join(['', self.name, ram.name, port, 'write_op_sel']),
            self.op_sel_width, initval=0)
        write_local_addr = self.m.Reg(
            '_'.join(['', self.name, ram.name, port, 'write_local_addr']),
            self.addrwidth, initval=0)
        write_size = self.m.Reg(
            '_'.join(['', self.name, ram.name, port, 'write_size']),
            self.addrwidth + 1, initval=0)
        write_local_stride = self.m.Reg(
            '_'.join(['', self.name, ram.name, port, 'write_local_stride']),
            self.addrwidth, initval=0)

        self.seq(
            write_start(0)
        )
        self.seq.If(start)(
            write_start(1),
            write_op_sel(op_id),
            write_local_addr(local_addr),
            write_size(size),
            write_local_stride(local_stride)
        )

        self.write_reqs[op_id] = (write_start, write_op_sel,
                                  write_local_addr,
                                  write_size, write_local_stride)

        if self.num_cmd_delay > 0:
            write_start = self.seq.Prev(write_start, self.num_cmd_delay)
            write_op_sel = self.seq.Prev(write_op_sel, self.num_cmd_delay)
            write_local_addr = self.seq.Prev(
                write_local_addr, self.num_cmd_delay)
            write_size = self.seq.Prev(write_size, self.num_cmd_delay)
            write_local_stride = self.seq.Prev(
                write_local_stride, self.num_cmd_delay)

        self.seq.If(write_start)(
            self.write_idle(0)
        )

        self.seq.If(write_start)(
            self.write_start(1),
            self.write_op_sel(write_op_sel),
            self.write_local_addr(write_local_addr),
            self.write_size(write_size),
            self.write_local_stride(write_local_stride)
        )

    def _synthesize_write_fsm(self, ram, port, ram_method):

        ram_method_name = (ram_method.func.__name__
                           if isinstance(ram_method, functools.partial) else
                           ram_method.__name__)
        ram_datawidth = (ram.datawidth if ram_method is None else
                         ram.orig_datawidth if 'bcast' in ram_method_name else
                         ram.orig_datawidth if 'block' in ram_method_name else
                         ram.datawidth)

        if not isinstance(self.datawidth, int):
            raise TypeError("axi.datawidth must be int, not '%s'" %
                            str(type(self.datawidth)))

        if not isinstance(ram_datawidth, int):
            raise TypeError("ram_datawidth must be int, not '%s'" %
                            str(type(ram_datawidth)))

        if self.datawidth == ram_datawidth:
            return self._synthesize_write_fsm_same(ram, port, ram_method, ram_datawidth)

        if self.datawidth < ram_datawidth:
            return self._synthesize_write_fsm_narrow(ram, port, ram_method, ram_datawidth)

        return self._synthesize_write_fsm_wide(ram, port, ram_method, ram_datawidth)

    def _synthesize_write_fsm_same(self, ram, port, ram_method, ram_datawidth):

        op_id = self._get_write_op_id(ram, port, ram_method)
        port = vtypes.to_int(port)

        if op_id in self.write_ops:
            """ already synthesized op """
            return

        if self.write_fsm is not None:
            """ new op """
            self.write_ops.append(op_id)

            fsm = self.write_fsm
            counter = self.write_counter

            # state 0
            fsm.set_index(0)
            cond = vtypes.Ands(self.write_start, self.write_op_sel == op_id)
            data, last, done = ram_method(
                port, self.write_local_addr, self.write_size,
                stride=self.write_local_stride, cond=cond, signed=False)

            if self.num_data_delay > 0:
                for _ in range(self.num_data_delay):
                    data = self.df._Delay(data)
                    last = self.df._Delay(last)

            fsm.If(cond).goto_next()

            # state 1
            fsm.set_index(1)
            cond = vtypes.Ands(fsm.here, self.write_op_sel == op_id)

            write_ready = vtypes.Ors(self.tdata.tready, vtypes.Not(self.tdata.tvalid))
            read_cond = vtypes.Ands(cond, write_ready)

            rdata, rvalid = data.read(cond=read_cond)
            rlast, _ = last.read(cond=read_cond)

            _ = self.write_data(rdata, rlast, cond=rvalid)

            return

        """ new op and fsm """
        fsm = FSM(self.m, '_'.join(['', self.name, 'write_fsm']),
                  self.clk, self.rst, as_module=self.fsm_as_module)
        self.write_fsm = fsm

        self.write_ops.append(op_id)

        # state 0
        cond = vtypes.Ands(self.write_start, self.write_op_sel == op_id)
        data, last, done = ram_method(
            port, self.write_local_addr, self.write_size,
            stride=self.write_local_stride, cond=cond, signed=False)

        if self.num_data_delay > 0:
            for _ in range(self.num_data_delay):
                data = self.df._Delay(data)
                last = self.df._Delay(last)

        counter = self.m.Reg('_'.join(['', self.name, 'write_counter']),
                             self.addrwidth + 1, initval=0)
        self.write_counter = counter

        fsm.If(self.write_start)(
            counter(self.write_size)
        )
        fsm.If(cond).goto_next()

        # state 1
        cond = vtypes.Ands(fsm.here, self.write_op_sel == op_id)

        write_ready = vtypes.Ors(self.tdata.tready, vtypes.Not(self.tdata.tvalid))
        read_cond = vtypes.Ands(cond, write_ready)

        rdata, rvalid = data.read(cond=read_cond)
        rlast, _ = last.read(cond=read_cond)

        _ = self.write_data(rdata, rlast, cond=rvalid)

        ack = vtypes.Ands(self.tdata.tvalid, self.tdata.tready)

        fsm.If(ack)(
            counter.dec()
        )

        done = counter == 0
        fsm.If(done).goto_next()

        # state 2
        set_idle = self._set_flag(fsm)
        self.seq.If(set_idle)(
            self.write_idle(1)
        )

        fsm.goto_init()

    def _synthesize_write_fsm_narrow(self, ram, port, ram_method, ram_datawidth):
        """ axi.datawidth < ram.datawidth """

        if ram_datawidth % self.datawidth != 0:
            raise ValueError(
                'ram_datawidth must be multiple number of axi.datawidth')

        pack_size = ram_datawidth // self.datawidth
        dma_size = (self.write_size << int(math.log(pack_size, 2))
                    if math.log(pack_size, 2) % 1.0 == 0.0 else
                    self.write_size * pack_size)

        op_id = self._get_write_op_id(ram, port, ram_method)
        port = vtypes.to_int(port)

        if op_id in self.write_ops:
            """ already synthesized op """
            return

        if pack_size in self.write_narrow_fsms:
            """ new op """
            self.write_ops.append(op_id)

            fsm = self.write_narrow_fsms[pack_size]
            counter = self.write_narrow_counters[pack_size]
            wdata = self.write_narrow_wdatas[pack_size]
            wvalid = self.write_narrow_wvalids[pack_size]
            wlast = self.write_narrow_wlasts[pack_size]
            pack_count = self.write_narrow_pack_counts[pack_size]

            # state 0
            fsm.set_index(0)
            cond = vtypes.Ands(self.write_start, self.write_op_sel == op_id)
            data, last, done = ram_method(
                port, self.write_local_addr, self.write_size,
                stride=self.write_local_stride, cond=cond, signed=False)

            if self.num_data_delay > 0:
                for _ in range(self.num_data_delay):
                    data = self.df._Delay(data)
                    last = self.df._Delay(last)

            fsm.If(cond).goto_next()

            # state 1
            fsm.set_index(1)
            cond = vtypes.Ands(fsm.here, self.write_op_sel == op_id)

            write_ready = vtypes.Ors(self.tdata.tready, vtypes.Not(self.tdata.tvalid))
            read_cond = vtyeps.Ands(cond, write_ready, pack_count == 0)

            rdata, rvalid = data.read(cond=read_cond)
            rlast, _ = last.read(cond=read_cond)

            stay_cond = self.write_op_sel == op_id

            self.seq.If(write_ready, rvalid, stay_cond)(
                wdata(rdata),
                wvalid(1),
                wlast(rlast),
                pack_count.inc()
            )

            return

        """ new op and fsm """
        fsm = FSM(self.m, '_'.join(['', self.name,
                                    'write_narrow', str(pack_size),
                                    'fsm']),
                  self.clk, self.rst, as_module=self.fsm_as_module)
        self.write_narrow_fsms[pack_size] = fsm

        self.write_ops.append(op_id)

        # state 0
        cond = vtypes.Ands(self.write_start, self.write_op_sel == op_id)
        data, last, done = ram_method(
            port, self.write_local_addr, self.write_size,
            stride=self.write_local_stride, cond=cond, signed=False)

        if self.num_data_delay > 0:
            for _ in range(self.num_data_delay):
                data = self.df._Delay(data)
                last = self.df._Delay(last)

        counter = self.m.Reg('_'.join(['', self.name,
                                       'write_narrow', str(pack_size), 'counter']),
                             self.addrwidth + 1, initval=0)
        self.write_narrow_counters[pack_size] = counter

        fsm.If(self.write_start)(
            counter(dma_size)
        )
        fsm.If(cond).goto_next()

        # state 1
        wdata = self.m.Reg('_'.join(['', self.name,
                                     'write_narrow', str(pack_size),
                                     'wdata']),
                           ram_datawidth, initval=0)
        self.write_narrow_wdatas[pack_size] = wdata
        wvalid = self.m.Reg('_'.join(['', self.name,
                                      'write_narrow', str(pack_size),
                                      'wvalid']),
                            initval=0)
        self.write_narrow_wvalids[pack_size] = wvalid
        wlast = self.m.Reg('_'.join(['', self.name,
                                     'write_narrow', str(pack_size),
                                     'wlast']),
                           initval=0)
        self.write_narrow_wlasts[pack_size] = wlast
        pack_count = self.m.Reg('_'.join(['', self.name,
                                          'write_narrow', str(pack_size),
                                          'pack_count']),
                                int(math.ceil(math.log(pack_size, 2))), initval=0)
        self.write_narrow_pack_counts[pack_size] = pack_count

        cond = vtypes.Ands(fsm.here, self.write_op_sel == op_id)

        write_ready = vtypes.Ors(self.tdata.tready, vtypes.Not(self.tdata.tvalid))
        read_cond = vtypes.Ands(cond, write_ready, pack_count == 0)

        rdata, rvalid = data.read(cond=read_cond)
        rlast, _ = last.read(cond=read_cond)

        stay_cond = self.write_op_sel == op_id

        self.seq.If(write_ready)(
            wvalid(0),
            wlast(0)
        )
        self.seq.If(write_ready, rvalid, stay_cond)(
            wdata(rdata),
            wvalid(1),
            wlast(rlast),
            pack_count.inc()
        )
        self.seq.If(write_ready, pack_count > 0)(
            wdata(wdata >> self.datawidth),
            wvalid(1),
            pack_count.inc()
        )
        self.seq.If(write_ready, pack_count == pack_size - 1)(
            wdata(wdata >> self.datawidth),
            wvalid(1),
            pack_count(0)
        )

        _ = self.write_data(wdata, wlast, cond=wvalid)

        ack = vtypes.Ands(self.tdata.tvalid, self.tdata.tready)
        fsm.If(ack)(
            counter.dec()
        )

        done = counter == 0
        fsm.If(done).goto_next()

        # state 2
        set_idle = self._set_flag(fsm)
        self.seq.If(set_idle)(
            self.write_idle(1)
        )

        fsm.goto_init()

    def _synthesize_write_fsm_wide(self, ram, port, ram_method, ram_datawidth):
        """ axi.datawidth > ram.datawidth """

        if self.datawidth % ram_datawidth != 0:
            raise ValueError(
                'axi.datawidth must be multiple number of ram_datawidth')

        pack_size = self.datawidth // ram_datawidth
        shamt = int(math.log(pack_size, 2))
        res = vtypes.Mux(
            vtypes.And(self.write_size, 2 ** shamt - 1) > 0, 1, 0)
        dma_size = (self.write_size >> shamt) + res

        actual_write_size = dma_size << shamt

        op_id = self._get_write_op_id(ram, port, ram_method)
        port = vtypes.to_int(port)

        if op_id in self.write_ops:
            """ already synthesized op """
            return

        if pack_size in self.write_wide_fsms:
            """ new op """
            self.write_ops.append(op_id)

            fsm = self.write_wide_fsms[pack_size]
            counter = self.write_wide_counters[pack_size]
            wdata = self.write_wide_wdatas[pack_size]
            wvalid = self.write_wide_wvalids[pack_size]
            wlast = self.write_wide_wlasts[pack_size]
            pack_count = self.write_wide_pack_counts[pack_size]

            # state 0
            fsm.set_index(0)
            cond = vtypes.Ands(self.write_start, self.write_op_sel == op_id)
            data, last, done = ram_method(
                port, self.write_local_addr, actual_write_size,
                stride=self.write_local_stride, cond=cond, signed=False)

            if self.num_data_delay > 0:
                for _ in range(self.num_data_delay):
                    data = self.df._Delay(data)
                    last = self.df._Delay(last)

            fsm.If(cond).goto_next()

            # state 1
            fsm.set_index(1)
            cond = vtypes.Ands(fsm.here, self.write_op_sel == op_id)

            write_ready = vtypes.Ors(self.tdata.tready, vtypes.Not(self.tdata.tvalid))
            read_cond = cond

            rdata, rvalid = data.read(cond=read_cond)
            rlast, _ = last.read(cond=read_cond)

            stay_cond = self.write_op_sel == op_id

            ack = vtypes.Ors(wready, vtypes.Not(wvalid))
            cond = vtypes.Ands(fsm.here, ack, self.write_op_sel == op_id)
            rdata, rvalid = data.read(cond=cond)

            self.seq.If(rvalid, stay_cond)(
                wdata(vtypes.Cat(rdata, wdata[ram_datawidth:self.datawidth])),
                wvalid(0),
                wlast(0),
                pack_count.inc()
            )
            self.seq.If(rvalid, stay_cond, pack_count == pack_size - 1)(
                wdata(vtypes.Cat(rdata, wdata[ram_datawidth:self.datawidth])),
                wvalid(1),
                wlast(rlast),
                pack_count(0)
            )

            return

        """ new op and fsm """
        fsm = FSM(self.m, '_'.join(['', self.name,
                                    'write_wide', str(pack_size),
                                    'fsm']),
                  self.clk, self.rst, as_module=self.fsm_as_module)
        self.write_wide_fsms[pack_size] = fsm

        self.write_ops.append(op_id)

        # state 0
        cond = vtypes.Ands(self.write_start, self.write_op_sel == op_id)
        data, last, done = ram_method(
            port, self.write_local_addr, actual_write_size,
            stride=self.write_local_stride, cond=cond, signed=False)

        if self.num_data_delay > 0:
            for _ in range(self.num_data_delay):
                data = self.df._Delay(data)
                last = self.df._Delay(last)

        counter = self.m.Reg('_'.join(['', self.name,
                                       'write_wide', str(pack_size), 'counter']),
                             self.addrwidth + 1, initval=0)
        self.write_wide_counters[pack_size] = counter

        fsm.If(self.write_start)(
            counter(dma_size)
        )
        fsm.If(cond).goto_next()

        # state 1
        wdata = self.m.Reg('_'.join(['', self.name,
                                     'write_wide', str(pack_size),
                                     'wdata']),
                           self.datawidth, initval=0)
        self.write_wide_wdatas[pack_size] = wdata
        wvalid = self.m.Reg('_'.join(['', self.name,
                                      'write_wide', str(pack_size),
                                      'wvalid']),
                            initval=0)
        self.write_wide_wvalids[pack_size] = wvalid
        wlast = self.m.Reg('_'.join(['', self.name,
                                     'write_wide', str(pack_size),
                                     'wlast']),
                           initval=0)
        self.write_wide_wlasts[pack_size] = wlast
        pack_count = self.m.Reg('_'.join(['', self.name,
                                          'write_wide', str(pack_size),
                                          'pack_count']),
                                int(math.ceil(math.log(pack_size, 2))), initval=0)
        self.write_wide_pack_counts[pack_size] = pack_count

        cond = vtypes.Ands(fsm.here, self.write_op_sel == op_id)

        write_ready = vtypes.Ors(self.tdata.tready, vtypes.Not(self.tdata.tvalid))
        read_cond = cond

        rdata, rvalid = data.read(cond=read_cond)
        rlast, _ = last.read(cond=read_cond)

        stay_cond = self.write_op_sel == op_id

        self.seq.If(write_ready)(
            wvalid(0),
            wlast(0)
        )
        self.seq.If(rvalid, stay_cond)(
            wdata(vtypes.Cat(rdata, wdata[ram_datawidth:self.datawidth])),
            wvalid(0),
            wlast(0),
            pack_count.inc()
        )
        self.seq.If(rvalid, stay_cond, pack_count == pack_size - 1)(
            wdata(vtypes.Cat(rdata, wdata[ram_datawidth:self.datawidth])),
            wvalid(1),
            wlast(rlast),
            pack_count(0)
        )

        _ = self.write_data(wdata, wlast, cond=wvalid)

        ack = vtypes.Ands(self.tdata.tvalid, self.tdata.tready)
        fsm.If(ack)(
            counter.dec()
        )

        done = counter == 0
        fsm.If(done).goto_next()

        # state 2
        set_idle = self._set_flag(fsm)
        self.seq.If(set_idle)(
            self.write_idle(1)
        )

        fsm.goto_init()

    def _set_flag(self, fsm, prefix='axistreamout_flag'):
        flag = self.m.TmpReg(initval=0, prefix=prefix)
        fsm(
            flag(1)
        )
        fsm.Delay(1)(
            flag(0)
        )
        fsm.goto_next()
        return flag

    def _get_write_op_id(self, ram, port, ram_method):

        ram_id = ram._id()
        port = vtypes.to_int(port)
        ram_method_name = (ram_method.func.__name__
                           if isinstance(ram_method, functools.partial) else
                           ram_method.__name__)
        op = (ram_id, port, ram_method_name)

        if op in self.write_op_id_map:
            op_id = self.write_op_id_map[op]
        else:
            op_id = self.write_op_id_count
            self.write_op_id_count += 1
            self.write_op_id_map[op] = op_id

        return op_id


class AXIStreamOutFifo(AXIStreamOut):
    """ AXI Stream Interface from FIFO for Output """

    __intrinsics__ = ('write',
                      'read_fifo',
                      'wait_read_fifo')

    def read_ram(self, fsm, ram, local_addr, size,
                 local_stride=1, port=0, ram_method=None):

        raise NotImplementedError('Use AXIStreamOut.')

    def read_ram_async(self, fsm, ram, local_addr, size,
                       local_stride=1, port=0, ram_method=None):

        raise NotImplementedError('Use AXIStreamOut.')

    def wait_read_ram(self, fsm):

        raise NotImplementedError('Use AXIStreamOut.')

    def read_fifo(self, fsm, fifo, size):

        if not self.enable_async:
            raise ValueError(
                "Async mode is disabled. Set 'True' to AXIM.enable_async.")

        self.wait_read_fifo(fsm)

        self._read_fifo(fsm, fifo, size)

    def wait_read_fifo(self, fsm):

        fsm.If(self.write_idle).goto_next()

    def _get_write_op_id_fifo(self, fifo):

        fifo_id = fifo._id()
        op = fifo_id

        if op in self.write_op_id_map:
            op_id = self.write_op_id_map[op]
        else:
            op_id = self.write_op_id_count
            self.write_op_id_count += 1
            self.write_op_id_map[op] = op_id

        return op_id

    def _read_fifo(self, fsm, fifo, size):

        if self.num_data_delay != 0:
            raise ValueError('num_data_delay must be 0.')

        if not isinstance(fifo, FIFO):
            raise TypeError('FIFO object is required.')

        start = self._set_flag(fsm)

        for _ in range(self.num_cmd_delay + 1):
            fsm.goto_next()

        self._set_write_request_fifo(fifo, start, size)

        self._synthesize_write_fsm_fifo(fifo)

        fsm.goto_next()

    def _set_write_request_fifo(self, fifo, start, size):

        op_id = self._get_write_op_id_fifo(fifo)

        if op_id in self.write_ops:
            (write_start, write_op_sel, write_size_in) = self.write_ops[op_id]

            self.seq.If(start)(
                write_start(1),
                write_op_sel(op_id)
            )

            return

        write_start = self.m.Reg(
            '_'.join(['', self.name, fifo.name, 'write_start']),
            initval=0)
        write_op_sel = self.m.Reg(
            '_'.join(['', self.name, fifo.name, 'write_op_sel']),
            self.op_sel_width, initval=0)
        write_size = self.m.Reg(
            '_'.join(['', self.name, fifo.name, 'write_size']),
            self.addrwidth + 1, initval=0)

        self.seq(
            write_start(0)
        )
        self.seq.If(start)(
            write_start(1),
            write_op_sel(op_id),
            write_size(size),
        )

        self.write_reqs[op_id] = (write_start, write_op_sel, write_size)

        if self.num_cmd_delay > 0:
            write_start = self.seq.Prev(write_start, self.num_cmd_delay)
            write_op_sel = self.seq.Prev(write_op_sel, self.num_cmd_delay)
            write_size = self.seq.Prev(write_size, self.num_cmd_delay)

        self.seq.If(write_start)(
            self.write_idle(0)
        )

        self.seq.If(write_start)(
            self.write_start(1),
            self.write_op_sel(write_op_sel),
            self.write_size(write_size),
        )

    def _synthesize_write_fsm_fifo(self, fifo):

        fifo_datawidth = fifo.datawidth

        if not isinstance(fifo_datawidth, int):
            raise TypeError("fifo_datawidth must be int, not '%s'" %
                            str(type(fifo_datawidth)))

        if self.datawidth == fifo_datawidth:
            return self._synthesize_write_fsm_fifo_same(fifo, fifo_datawidth)

        if self.datawidth < fifo_datawidth:
            return self._synthesize_write_fsm_fifo_narrow(fifo, fifo_datawidth)

        return self._synthesize_write_fsm_fifo_wide(fifo, fifo_datawidth)

    def _synthesize_write_fsm_fifo_same(self, fifo, fifo_datawidth):

        op_id = self._get_write_op_id_fifo(fifo)

        if op_id in self.write_ops:
            """ already synthesized op """
            return

        if self.write_fsm is not None:
            """ new op """
            self.write_ops.append(op_id)

            fsm = self.write_fsm
            counter = self.write_counter

            # state 0
            fsm.set_index(0)
            cond = vtypes.Ands(self.write_start, self.write_op_sel == op_id)

            fifo_counter = self.m.TmpReg(prefix='_'.join(['', self.name, 'write_fifo_counter']),
                                         width=self.addrwidth + 1, initval=0)
            fsm.If(self.write_start)(
                fifo_counter(self.write_size)
            )

            fsm.If(cond).goto_next()

            # state 1
            fsm.set_index(1)
            cond = vtypes.Ands(fsm.here, self.write_op_sel == op_id)
            write_ready = vtypes.Ors(self.tdata.tready, vtypes.Not(self.tdata.tvalid))
            fifo_ready = vtypes.Not(fifo.empty)
            deq_cond = vtypes.Ands(cond, write_ready, fifo_ready, fifo_counter > 0)

            rdata, rvalid, _ = fifo.deq_rtl(cond=deq_cond)
            rlast = self.m.TmpReg(prefix='rlast', initval=0)
            self.seq.If(deq_cond)(
                rlast(fifo_counter <= 1)
            )
            fsm.If(deq_cond)(
                fifo_counter.dec()
            )

            repeat_rvalid = self.m.TmpReg(prefix='repeat_rvalid', initval=0)
            self.seq(repeat_rvalid(0))
            self.seq.If(rvalid, vtypes.Not(write_ready))(repeat_rvalid(1))
            self.seq.If(repeat_rvalid, vtypes.Not(write_ready))(repeat_rvalid(1))
            cur_rvalid = vtypes.Ors(rvalid, repeat_rvalid)

            _ = self.write_data(rdata, rlast, cond=cur_rvalid)

            return

        """ new op and fsm """
        fsm = FSM(self.m, '_'.join(['', self.name, 'write_fsm']),
                  self.clk, self.rst, as_module=self.fsm_as_module)
        self.write_fsm = fsm

        self.write_ops.append(op_id)

        # state 0
        cond = vtypes.Ands(self.write_start, self.write_op_sel == op_id)

        counter = self.m.Reg('_'.join(['', self.name, 'write_counter']),
                             self.addrwidth + 1, initval=0)
        self.write_counter = counter

        fsm.If(self.write_start)(
            counter(self.write_size)
        )

        fifo_counter = self.m.TmpReg(prefix='_'.join(['', self.name, 'write_fifo_counter']),
                                     width=self.addrwidth + 1, initval=0)
        fsm.If(self.write_start)(
            fifo_counter(self.write_size)
        )

        fsm.If(cond).goto_next()

        # state 1
        cond = vtypes.Ands(fsm.here, self.write_op_sel == op_id)
        write_ready = vtypes.Ors(self.tdata.tready, vtypes.Not(self.tdata.tvalid))
        fifo_ready = vtypes.Not(fifo.empty)
        deq_cond = vtypes.Ands(cond, write_ready, fifo_ready, fifo_counter > 0)

        rdata, rvalid, _ = fifo.deq_rtl(cond=deq_cond)
        rlast = self.m.TmpReg(prefix='rlast', initval=0)
        self.seq.If(deq_cond)(
            rlast(fifo_counter <= 1)
        )
        fsm.If(deq_cond)(
            fifo_counter.dec()
        )

        repeat_rvalid = self.m.TmpReg(prefix='repeat_rvalid', initval=0)
        self.seq(repeat_rvalid(0))
        self.seq.If(rvalid, vtypes.Not(write_ready))(repeat_rvalid(1))
        self.seq.If(repeat_rvalid, vtypes.Not(write_ready))(repeat_rvalid(1))
        cur_rvalid = vtypes.Ors(rvalid, repeat_rvalid)

        _ = self.write_data(rdata, rlast, cond=cur_rvalid)

        ack = vtypes.Ands(self.tdata.tvalid, self.tdata.tready)

        fsm.If(ack)(
            counter.dec()
        )

        done = counter == 0
        fsm.If(done).goto_next()

        # state 2
        set_idle = self._set_flag(fsm)
        self.seq.If(set_idle)(
            self.write_idle(1)
        )

        fsm.goto_init()

    def _synthesize_write_fsm_fifo_narrow(self, fifo, fifo_datawidth):
        """ axi.datawidth < fifo.datawidth """

        if fifo_datawidth % self.datawidth != 0:
            raise ValueError(
                'fifo_datawidth must be multiple number of axi.datawidth')

        pack_size = fifo_datawidth // self.datawidth
        dma_size = (self.write_size << int(math.log(pack_size, 2))
                    if math.log(pack_size, 2) % 1.0 == 0.0 else
                    self.write_size * pack_size)

        op_id = self._get_write_op_id_fifo(fifo)

        if op_id in self.write_ops:
            """ already synthesized op """
            return

        if pack_size in self.write_narrow_fsms:
            """ new op """
            self.write_ops.append(op_id)

            fsm = self.write_narrow_fsms[pack_size]
            counter = self.write_narrow_counters[pack_size]
            wdata = self.write_narrow_wdatas[pack_size]
            wvalid = self.write_narrow_wvalids[pack_size]
            wlast = self.write_narrow_wlasts[pack_size]
            pack_count = self.write_narrow_pack_counts[pack_size]

            # state 0
            fsm.set_index(0)
            cond = vtypes.Ands(self.write_start, self.write_op_sel == op_id)

            fifo_counter = self.m.TmpReg(prefix='_'.join(
                ['', self.name, 'write_narrow', str(pack_size), 'counter']),
                                         width=self.addrwidth + 1, initval=0)
            fsm.If(self.write_start)(
                fifo_counter(self.write_size)
            )

            fsm.If(cond).goto_next()

            # state 1
            fsm.set_index(1)
            cond = vtypes.Ands(fsm.here, self.write_op_sel == op_id)

            write_ready = vtypes.Ors(self.tdata.tready, vtypes.Not(self.tdata.tvalid))
            fifo_ready = vtypes.Not(fifo.empty)
            deq_cond = vtypes.Ands(cond, write_ready, fifo_ready,
                                   vtypes.Ors(pack_count == pack_size - 1,
                                              vtypes.Ands(pack_count == 0,
                                                          fifo_counter == self.write_size)),
                                   fifo_counter > 0)

            rdata, rvalid, _ = fifo.deq_rtl(cond=deq_cond)
            rlast = self.m.TmpReg(prefix='rlast', initval=0)
            self.seq.If(deq_cond)(
                rlast(fifo_counter <= 1)
            )
            fsm.If(deq_cond)(
                fifo_counter.dec()
            )

            repeat_rvalid = self.m.TmpReg(prefix='repeat_rvalid', initval=0)
            self.seq(repeat_rvalid(0))
            self.seq.If(rvalid, vtypes.Not(write_ready))(repeat_rvalid(1))
            self.seq.If(repeat_rvalid, vtypes.Not(write_ready))(repeat_rvalid(1))
            cur_rvalid = vtypes.Ors(rvalid, repeat_rvalid)

            stay_cond = self.write_op_sel == op_id

            self.seq.If(write_ready, cur_rvalid, stay_cond)(
                wdata(rdata),
                wvalid(1),
                wlast(0),
                pack_count.inc()
            )
            self.seq.If(write_ready, pack_count == pack_size - 1)(
                wdata(wdata >> self.datawidth),
                wvalid(1),
                wlast(rlast),
                pack_count(0)
            )

            return

        """ new op and fsm """
        fsm = FSM(self.m, '_'.join(['', self.name,
                                    'write_narrow', str(pack_size),
                                    'fsm']),
                  self.clk, self.rst, as_module=self.fsm_as_module)
        self.write_narrow_fsms[pack_size] = fsm

        self.write_ops.append(op_id)

        # state 0
        cond = vtypes.Ands(self.write_start, self.write_op_sel == op_id)

        counter = self.m.Reg('_'.join(['', self.name,
                                       'write_narrow', str(pack_size), 'counter']),
                             self.addrwidth + 1, initval=0)
        self.write_narrow_counters[pack_size] = counter

        fsm.If(self.write_start)(
            counter(dma_size)
        )

        fifo_counter = self.m.TmpReg(prefix='_'.join(['', self.name,
                                                      'write_narrow', str(pack_size), 'counter']),
                                     width=self.addrwidth + 1, initval=0)
        fsm.If(self.write_start)(
            fifo_counter(self.write_size)
        )

        fsm.If(cond).goto_next()

        # state 1
        wdata = self.m.Reg('_'.join(['', self.name,
                                     'write_narrow', str(pack_size),
                                     'wdata']),
                           fifo_datawidth, initval=0)
        self.write_narrow_wdatas[pack_size] = wdata
        wvalid = self.m.Reg('_'.join(['', self.name,
                                      'write_narrow', str(pack_size),
                                      'wvalid']),
                            initval=0)
        self.write_narrow_wvalids[pack_size] = wvalid
        wlast = self.m.Reg('_'.join(['', self.name,
                                     'write_narrow', str(pack_size),
                                     'wlast']),
                           initval=0)
        self.write_narrow_wlasts[pack_size] = wlast
        pack_count = self.m.Reg('_'.join(['', self.name,
                                          'write_narrow', str(pack_size),
                                          'pack_count']),
                                int(math.ceil(math.log(pack_size, 2))), initval=0)
        self.write_narrow_pack_counts[pack_size] = pack_count

        cond = vtypes.Ands(fsm.here, self.write_op_sel == op_id)

        write_ready = vtypes.Ors(self.tdata.tready, vtypes.Not(self.tdata.tvalid))
        fifo_ready = vtypes.Not(fifo.empty)
        deq_cond = vtypes.Ands(cond, write_ready, fifo_ready,
                               vtypes.Ors(pack_count == pack_size - 1,
                                          vtypes.Ands(pack_count == 0,
                                                      fifo_counter == self.write_size)),
                               fifo_counter > 0)

        rdata, rvalid, _ = fifo.deq_rtl(cond=deq_cond)
        rlast = self.m.TmpReg(prefix='rlast', initval=0)
        self.seq.If(deq_cond)(
            rlast(fifo_counter <= 1)
        )
        fsm.If(deq_cond)(
            fifo_counter.dec()
        )

        repeat_rvalid = self.m.TmpReg(prefix='repeat_rvalid', initval=0)
        self.seq(repeat_rvalid(0))
        self.seq.If(rvalid, vtypes.Not(write_ready))(repeat_rvalid(1))
        self.seq.If(repeat_rvalid, vtypes.Not(write_ready))(repeat_rvalid(1))
        cur_rvalid = vtypes.Ors(rvalid, repeat_rvalid)

        self.m.TmpWireLike(rdata, prefix='debug_rdata').assign(rdata)
        self.m.TmpWireLike(rvalid, prefix='debug_rvalid').assign(rvalid)

        stay_cond = self.write_op_sel == op_id

        self.seq.If(write_ready)(
            wvalid(0),
            wlast(0)
        )
        self.seq.If(write_ready, cur_rvalid, stay_cond)(
            wdata(rdata),
            wvalid(1),
            wlast(0),
            pack_count.inc()
        )
        self.seq.If(write_ready, pack_count > 0)(
            wdata(wdata >> self.datawidth),
            wvalid(1),
            wlast(0),
            pack_count.inc()
        )
        self.seq.If(write_ready, pack_count == pack_size - 1)(
            wdata(wdata >> self.datawidth),
            wvalid(1),
            wlast(rlast),
            pack_count(0)
        )

        _ = self.write_data(wdata, wlast, cond=wvalid)

        ack = vtypes.Ands(self.tdata.tvalid, self.tdata.tready)
        fsm.If(ack)(
            counter.dec()
        )

        done = counter == 0
        fsm.If(done).goto_next()

        # state 2
        set_idle = self._set_flag(fsm)
        self.seq.If(set_idle)(
            self.write_idle(1)
        )

        fsm.goto_init()

    def _synthesize_write_fsm_fifo_wide(self, fifo, fifo_datawidth):
        """ axi.datawidth > fifo.datawidth """

        if self.datawidth % fifo_datawidth != 0:
            raise ValueError(
                'axi.datawidth must be multiple number of fifo_datawidth')

        pack_size = self.datawidth // fifo_datawidth
        shamt = int(math.log(pack_size, 2))
        res = vtypes.Mux(
            vtypes.And(self.write_size, 2 ** shamt - 1) > 0, 1, 0)
        dma_size = (self.write_size >> shamt) + res

        actual_write_size = dma_size << shamt

        op_id = self._get_write_op_id_fifo(fifo)

        if op_id in self.write_ops:
            """ already synthesized op """
            return

        if pack_size in self.write_wide_fsms:
            """ new op """
            self.write_ops.append(op_id)

            fsm = self.write_narrow_fsms[pack_size]
            counter = self.write_narrow_counters[pack_size]
            wdata = self.write_narrow_wdatas[pack_size]
            wvalid = self.write_narrow_wvalids[pack_size]
            wlast = self.write_narrow_wlasts[pack_size]
            pack_count = self.write_narrow_pack_counts[pack_size]

            # state 0
            fsm.set_index(0)
            cond = vtypes.Ands(self.write_start, self.write_op_sel == op_id)

            fsm.If(cond).goto_next()

            # state 1
            fsm.set_index(1)
            cond = vtypes.Ands(fsm.here, self.write_op_sel == op_id)

            write_ready = vtypes.Ors(self.tdata.tready, vtypes.Not(self.tdata.tvalid))
            fifo_ready = vtypes.Not(fifo.empty)
            deq_cond = vtypes.Ands(cond, fifo_ready, counter > 0)

            rdata, rvalid, _ = fifo.deq_rtl(cond=deq_cond)
            rlast = counter <= 1

            repeat_rvalid = self.m.TmpReg(prefix='repeat_rvalid', initval=0)
            self.seq(repeat_rvalid(0))
            self.seq.If(rvalid, vtypes.Not(write_ready))(repeat_rvalid(1))
            self.seq.If(repeat_rvalid, vtypes.Not(write_ready))(repeat_rvalid(1))
            cur_rvalid = vtypes.Ors(rvalid, repeat_rvalid)

            stay_cond = self.write_op_sel == op_id

            self.seq.If(write_ready)(
                wvalid(0),
                wlast(0)
            )
            self.seq.If(cur_rvalid, stay_cond)(
                wdata(vtypes.Cat(rdata, wdata[fifo_datawidth:self.datawidth])),
                wvalid(0),
                wlast(0),
                pack_count.inc()
            )
            self.seq.If(cur_rvalid, stay_cond, pack_count == pack_size - 1)(
                wdata(vtypes.Cat(rdata, wdata[fifo_datawidth:self.datawidth])),
                wvalid(1),
                wlast(rlast),
                pack_count(0)
            )

            return

        """ new op and fsm """
        fsm = FSM(self.m, '_'.join(['', self.name,
                                    'write_wide', str(pack_size),
                                    'fsm']),
                  self.clk, self.rst, as_module=self.fsm_as_module)
        self.write_wide_fsms[pack_size] = fsm

        self.write_ops.append(op_id)

        # state 0
        cond = vtypes.Ands(self.write_start, self.write_op_sel == op_id)

        counter = self.m.Reg('_'.join(['', self.name,
                                       'write_wide', str(pack_size), 'counter']),
                             self.addrwidth + 1, initval=0)
        self.write_wide_counters[pack_size] = counter

        fsm.If(self.write_start)(
            counter(dma_size)
        )
        fsm.If(cond).goto_next()

        # state 1
        wdata = self.m.Reg('_'.join(['', self.name,
                                     'write_wide', str(pack_size),
                                     'wdata']),
                           self.datawidth, initval=0)
        self.write_wide_wdatas[pack_size] = wdata
        wvalid = self.m.Reg('_'.join(['', self.name,
                                      'write_wide', str(pack_size),
                                      'wvalid']),
                            initval=0)
        self.write_wide_wvalids[pack_size] = wvalid
        wlast = self.m.Reg('_'.join(['', self.name,
                                     'write_wide', str(pack_size),
                                     'wlast']),
                           initval=0)
        self.write_wide_wlasts[pack_size] = wlast
        pack_count = self.m.Reg('_'.join(['', self.name,
                                          'write_wide', str(pack_size),
                                          'pack_count']),
                                int(math.ceil(math.log(pack_size, 2))), initval=0)
        self.write_wide_pack_counts[pack_size] = pack_count

        cond = vtypes.Ands(fsm.here, self.write_op_sel == op_id)

        write_ready = vtypes.Ors(self.tdata.tready, vtypes.Not(self.tdata.tvalid))
        fifo_ready = vtypes.Not(fifo.empty)
        deq_cond = vtypes.Ands(cond, fifo_ready, counter > 0)

        rdata, rvalid, _ = fifo.deq_rtl(cond=deq_cond)
        rlast = counter <= 1

        repeat_rvalid = self.m.TmpReg(prefix='repeat_rvalid', initval=0)
        self.seq(repeat_rvalid(0))
        self.seq.If(rvalid, vtypes.Not(write_ready))(repeat_rvalid(1))
        self.seq.If(repeat_rvalid, vtypes.Not(write_ready))(repeat_rvalid(1))
        cur_rvalid = vtypes.Ors(rvalid, repeat_rvalid)

        stay_cond = self.write_op_sel == op_id

        self.seq.If(write_ready)(
            wvalid(0),
            wlast(0)
        )
        self.seq.If(cur_rvalid, stay_cond)(
            wdata(vtypes.Cat(rdata, wdata[fifo_datawidth:self.datawidth])),
            wvalid(0),
            wlast(0),
            pack_count.inc()
        )
        self.seq.If(cur_rvalid, stay_cond, pack_count == pack_size - 1)(
            wdata(vtypes.Cat(rdata, wdata[fifo_datawidth:self.datawidth])),
            wvalid(1),
            wlast(rlast),
            pack_count(0)
        )

        _ = self.write_data(wdata, wlast, cond=wvalid)

        ack = vtypes.Ands(self.tdata.tvalid, self.tdata.tready)
        fsm.If(ack)(
            counter.dec()
        )

        done = counter == 0
        fsm.If(done).goto_next()

        # state 2
        set_idle = self._set_flag(fsm)
        self.seq.If(set_idle)(
            self.write_idle(1)
        )

        fsm.goto_init()
